# Copyright 2011 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
import six
import webob

import ddt
from oslo_utils import strutils

from cinder.api.contrib import types_manage
from cinder import context
from cinder import exception
from cinder import test
from cinder.tests.unit.api import fakes
from cinder.tests.unit import fake_constants as fake
from cinder.volume import volume_types

DEFAULT_VOLUME_TYPE = fake.VOLUME_TYPE_ID
IN_USE_VOLUME_TYPE = fake.VOLUME_TYPE2_ID
UPDATE_DESC_ONLY_TYPE = fake.VOLUME_TYPE3_ID
UPDATE_NAME_ONLY_TYPE = fake.VOLUME_TYPE4_ID
UPDATE_NAME_AFTER_DELETE_TYPE = fake.VOLUME_TYPE5_ID
NOT_FOUND_VOLUME_TYPE = fake.WILL_NOT_BE_FOUND_ID


def fake_volume_type(id):
    specs = {"key1": "value1",
             "key2": "value2",
             "key3": "value3",
             "key4": "value4",
             "key5": "value5"}
    return dict(id=id,
                name='vol_type_%s' % six.text_type(id),
                description='vol_type_desc_%s' % six.text_type(id),
                extra_specs=specs)


def fake_volume_type_updated(id, is_public=True):
    return dict(id=id,
                name='vol_type_%s_%s' % (six.text_type(id), six.text_type(id)),
                is_public=is_public,
                description='vol_type_desc_%s_%s' % (
                    six.text_type(id), six.text_type(id)))


def fake_volume_type_updated_desc_only(id):
    return dict(id=id,
                name='vol_type_%s' % six.text_type(id),
                description='vol_type_desc_%s_%s' % (
                    six.text_type(id), six.text_type(id)))


def return_volume_types_get_volume_type(context, id):
    if id == fake.WILL_NOT_BE_FOUND_ID:
        raise exception.VolumeTypeNotFound(volume_type_id=id)
    return fake_volume_type(id)


def return_volume_types_destroy(context, name):
    if name == fake.WILL_NOT_BE_FOUND_ID:
        raise exception.VolumeTypeNotFoundByName(volume_type_name=name)
    pass


def return_volume_types_with_volumes_destroy(context, id):
    if id == IN_USE_VOLUME_TYPE:
        raise exception.VolumeTypeInUse(volume_type_id=id)
    pass


def return_volume_types_create(context,
                               name,
                               specs,
                               is_public,
                               description):
    pass


def return_volume_types_create_duplicate_type(context,
                                              name,
                                              specs,
                                              is_public,
                                              description):
    raise exception.VolumeTypeExists(id=name)


def fake_volume_type_updated_name_only(id):
    return dict(id=id,
                name='vol_type_%s_%s' % (six.text_type(id), six.text_type(id)),
                description='vol_type_desc_%s' % six.text_type(id))


def fake_volume_type_updated_name_after_delete(id):
    return dict(id=id,
                name='vol_type_%s' % six.text_type(id),
                description='vol_type_desc_%s' % six.text_type(id))


def return_volume_types_get_volume_type_updated(id, is_public=True):
    if id == NOT_FOUND_VOLUME_TYPE:
        raise exception.VolumeTypeNotFound(volume_type_id=id)
    if id == UPDATE_DESC_ONLY_TYPE:
        return fake_volume_type_updated_desc_only(id)
    if id == UPDATE_NAME_ONLY_TYPE:
        return fake_volume_type_updated_name_only(id)
    if id == UPDATE_NAME_AFTER_DELETE_TYPE:
        return fake_volume_type_updated_name_after_delete(id)

    # anything else
    return fake_volume_type_updated(id, is_public=is_public)


def return_volume_types_get_by_name(context, name):
    if name == NOT_FOUND_VOLUME_TYPE:
        raise exception.VolumeTypeNotFoundByName(volume_type_name=name)
    return fake_volume_type(name.split("_")[2])


def return_volume_types_get_default():
    return fake_volume_type(DEFAULT_VOLUME_TYPE)


def return_volume_types_get_default_not_found():
    return {}


@ddt.ddt
class VolumeTypesManageApiTest(test.TestCase):
    def setUp(self):
        super(VolumeTypesManageApiTest, self).setUp()
        self.flags(host='fake')
        self.controller = types_manage.VolumeTypesManageController()
        """to reset notifier drivers left over from other api/contrib tests"""

    def test_volume_types_delete(self):
        self.mock_object(volume_types, 'get_volume_type',
                         return_volume_types_get_volume_type)
        self.mock_object(volume_types, 'destroy',
                         return_volume_types_destroy)

        req = fakes.HTTPRequest.blank('/v2/%s/types/%s' % (
            fake.PROJECT_ID, DEFAULT_VOLUME_TYPE))
        self.assertEqual(0, len(self.notifier.notifications))
        self.controller._delete(req, DEFAULT_VOLUME_TYPE)
        self.assertEqual(1, len(self.notifier.notifications))

    def test_volume_types_delete_not_found(self):
        self.mock_object(volume_types, 'get_volume_type',
                         return_volume_types_get_volume_type)
        self.mock_object(volume_types, 'destroy',
                         return_volume_types_destroy)

        self.assertEqual(0, len(self.notifier.notifications))
        req = fakes.HTTPRequest.blank('/v2/%s/types/%s' % (
            fake.PROJECT_ID, NOT_FOUND_VOLUME_TYPE))
        self.assertRaises(exception.VolumeTypeNotFound,
                          self.controller._delete, req, NOT_FOUND_VOLUME_TYPE)
        self.assertEqual(1, len(self.notifier.notifications))

    def test_volume_types_with_volumes_destroy(self):
        self.mock_object(volume_types, 'get_volume_type',
                         return_volume_types_get_volume_type)
        self.mock_object(volume_types, 'destroy',
                         return_volume_types_with_volumes_destroy)
        req = fakes.HTTPRequest.blank('/v2/%s/types/%s' % (
            fake.PROJECT_ID, DEFAULT_VOLUME_TYPE))
        self.assertEqual(0, len(self.notifier.notifications))
        self.controller._delete(req, DEFAULT_VOLUME_TYPE)
        self.assertEqual(1, len(self.notifier.notifications))

    @mock.patch('cinder.volume.volume_types.destroy')
    @mock.patch('cinder.volume.volume_types.get_volume_type')
    @mock.patch('cinder.policy.authorize')
    def test_volume_types_delete_with_non_admin(self, mock_policy_authorize,
                                                mock_get, mock_destroy):

        # allow policy authorized user to delete type
        mock_policy_authorize.return_value = None
        mock_get.return_value = \
            {'extra_specs': {"key1": "value1"},
             'id': DEFAULT_VOLUME_TYPE,
             'name': u'vol_type_1',
             'description': u'vol_type_desc_%s' % DEFAULT_VOLUME_TYPE}
        mock_destroy.side_effect = return_volume_types_destroy

        req = fakes.HTTPRequest.blank('/v2/%s/types/%s' %
                                      (fake.PROJECT_ID, DEFAULT_VOLUME_TYPE),
                                      use_admin_context=False)
        self.assertEqual(0, len(self.notifier.notifications))
        self.controller._delete(req, DEFAULT_VOLUME_TYPE)
        self.assertEqual(1, len(self.notifier.notifications))
        # non policy authorized user fails to delete type
        mock_policy_authorize.side_effect = (
            exception.PolicyNotAuthorized(action='type_delete'))
        self.assertRaises(exception.PolicyNotAuthorized,
                          self.controller._delete,
                          req, DEFAULT_VOLUME_TYPE)

    def test_create(self):
        self.mock_object(volume_types, 'create',
                         return_volume_types_create)
        self.mock_object(volume_types, 'get_volume_type_by_name',
                         return_volume_types_get_by_name)

        body = {"volume_type": {"name": "vol_type_1",
                                "os-volume-type-access:is_public": True,
                                "extra_specs": {"key1": "value1"}}}
        req = fakes.HTTPRequest.blank('/v2/%s/types' % fake.PROJECT_ID)

        self.assertEqual(0, len(self.notifier.notifications))
        res_dict = self.controller._create(req, body=body)

        self.assertEqual(1, len(self.notifier.notifications))
        id = res_dict['volume_type']['id']
        self._check_test_results(res_dict, {
            'expected_name': 'vol_type_1',
            'expected_desc': 'vol_type_desc_%s' % id})

    @mock.patch('cinder.volume.volume_types.create')
    @mock.patch('cinder.volume.volume_types.get_volume_type_by_name')
    def test_create_with_description_of_zero_length(
            self, mock_get_volume_type_by_name, mock_create_type):
        mock_get_volume_type_by_name.return_value = \
            {'extra_specs': {"key1": "value1"},
             'id': DEFAULT_VOLUME_TYPE,
             'name': u'vol_type_1',
             'description': u''}

        type_description = ""
        body = {"volume_type": {"name": "vol_type_1",
                                "description": type_description,
                                "extra_specs": {"key1": "value1"}}}
        req = fakes.HTTPRequest.blank('/v2/%s/types' % fake.PROJECT_ID)

        res_dict = self.controller._create(req, body=body)

        self._check_test_results(res_dict, {
            'expected_name': 'vol_type_1', 'expected_desc': ''})

    def test_create_type_with_name_too_long(self):
        type_name = 'a' * 256
        body = {"volume_type": {"name": type_name,
                                "extra_specs": {"key1": "value1"}}}
        req = fakes.HTTPRequest.blank('/v2/%s/types' % fake.PROJECT_ID)
        self.assertRaises(exception.ValidationError,
                          self.controller._create, req, body=body)

    def test_create_type_with_description_too_long(self):
        type_description = 'a' * 256
        body = {"volume_type": {"name": "vol_type_1",
                                "description": type_description,
                                "extra_specs": {"key1": "value1"}}}
        req = fakes.HTTPRequest.blank('/v2/%s/types' % fake.PROJECT_ID)
        self.assertRaises(exception.ValidationError,
                          self.controller._create, req, body=body)

    def test_create_duplicate_type_fail(self):
        self.mock_object(volume_types, 'create',
                         return_volume_types_create_duplicate_type)
        self.mock_object(volume_types, 'get_volume_type_by_name',
                         return_volume_types_get_by_name)

        body = {"volume_type": {"name": "vol_type_1",
                                "extra_specs": {"key1": "value1"}}}
        req = fakes.HTTPRequest.blank('/v2/%s/types' % fake.PROJECT_ID)
        self.assertRaises(webob.exc.HTTPConflict,
                          self.controller._create, req, body=body)

    def test_create_type_with_invalid_is_public(self):
        body = {"volume_type": {"name": "vol_type_1",
                                "os-volume-type-access:is_public": "fake",
                                "description": "test description",
                                "extra_specs": {"key1": "value1"}}}
        req = fakes.HTTPRequest.blank('/v2/%s/types' % fake.PROJECT_ID)
        self.assertRaises(exception.ValidationError,
                          self.controller._create, req, body=body)

    @ddt.data('0', 'f', 'false', 'off', 'n', 'no', '1', 't', 'true', 'on',
              'y', 'yes')
    @mock.patch.object(volume_types, "get_volume_type_by_name")
    @mock.patch.object(volume_types, "create")
    @mock.patch("cinder.api.openstack.wsgi.Request.cache_resource")
    @mock.patch("cinder.api.views.types.ViewBuilder.show")
    def test_create_type_with_valid_is_public_in_string(
            self, is_public, mock_show, mock_cache_resource,
            mock_create, mock_get):
        boolean_is_public = strutils.bool_from_string(is_public)
        ctxt = context.RequestContext(fake.USER_ID, fake.PROJECT_ID, True)
        body = {"volume_type": {"name": "vol_type_1",
                                "os-volume-type-access:is_public":
                                    is_public,
                                "extra_specs": {"key1": "value1"}}}
        req = fakes.HTTPRequest.blank('/v2/%s/types' % fake.PROJECT_ID)
        req.environ['cinder.context'] = ctxt
        self.controller._create(req, body=body)
        mock_create.assert_called_once_with(
            ctxt, 'vol_type_1', {'key1': 'value1'},
            boolean_is_public, description=None)

    def _create_volume_type_bad_body(self, body):
        req = fakes.HTTPRequest.blank('/v2/%s/types' % fake.PROJECT_ID)
        req.method = 'POST'
        self.assertRaises(exception.ValidationError,
                          self.controller._create, req, body=body)

    def test_create_no_body(self):
        self._create_volume_type_bad_body(body=None)

    def test_create_missing_volume(self):
        body = {'foo': {'a': 'b'}}
        self._create_volume_type_bad_body(body=body)

    def test_create_malformed_entity(self):
        body = {'volume_type': 'string'}
        self._create_volume_type_bad_body(body=body)

    @mock.patch('cinder.volume.volume_types.create')
    @mock.patch('cinder.volume.volume_types.get_volume_type_by_name')
    @mock.patch('cinder.policy.authorize')
    def test_create_with_none_admin(self, mock_policy_authorize,
                                    mock_get_volume_type_by_name,
                                    mock_create_type):

        # allow policy authorized user to create type
        mock_policy_authorize.return_value = None
        mock_get_volume_type_by_name.return_value = \
            {'extra_specs': {"key1": "value1"},
             'id': DEFAULT_VOLUME_TYPE,
             'name': u'vol_type_1',
             'description': u'vol_type_desc_1'}

        body = {"volume_type": {"name": "vol_type_1",
                                "os-volume-type-access:is_public": True,
                                "extra_specs": {"key1": "value1"}}}
        req = fakes.HTTPRequest.blank('/v2/%s/types' % fake.PROJECT_ID,
                                      use_admin_context=False)

        self.assertEqual(0, len(self.notifier.notifications))
        res_dict = self.controller._create(req, body=body)

        self.assertEqual(1, len(self.notifier.notifications))
        self._check_test_results(res_dict, {
            'expected_name': 'vol_type_1', 'expected_desc': 'vol_type_desc_1'})

        # non policy authorized user fails to create type
        mock_policy_authorize.side_effect = (
            exception.PolicyNotAuthorized(action='type_create'))
        self.assertRaises(exception.PolicyNotAuthorized,
                          self.controller._create,
                          req, body=body)

    @ddt.data({'a' * 256: 'a'},
              {'a': 'a' * 256},
              {'': 'a'},
              'foo',
              None)
    def test_create_type_with_invalid_extra_specs(self, value):
        body = {"volume_type": {"name": "vol_type_1",
                                "os-volume-type-access:is_public": False,
                                "description": "test description"}}
        body['volume_type']['extra_specs'] = value
        req = fakes.HTTPRequest.blank('/v2/%s/types' % fake.PROJECT_ID)
        self.assertRaises(exception.ValidationError,
                          self.controller._create, req, body=body)

    @mock.patch('cinder.volume.volume_types.update')
    @mock.patch('cinder.volume.volume_types.get_volume_type')
    def test_update(self, mock_get, mock_update):
        mock_get.return_value = return_volume_types_get_volume_type_updated(
            DEFAULT_VOLUME_TYPE, is_public=False)
        body = {"volume_type": {"is_public": False}}
        req = fakes.HTTPRequest.blank('/v2/%s/types/%s' % (
            fake.PROJECT_ID, DEFAULT_VOLUME_TYPE))
        req.method = 'PUT'

        self.assertEqual(0, len(self.notifier.notifications))
        res_dict = self.controller._update(req, DEFAULT_VOLUME_TYPE, body=body)
        self.assertEqual(1, len(self.notifier.notifications))
        self._check_test_results(
            res_dict,
            {'expected_desc': 'vol_type_desc_%s_%s' %
                              (DEFAULT_VOLUME_TYPE, DEFAULT_VOLUME_TYPE),
             'expected_name': 'vol_type_%s_%s' %
                              (DEFAULT_VOLUME_TYPE, DEFAULT_VOLUME_TYPE),
             'is_public': False})

    @ddt.data('0', 'f', 'false', 'off', 'n', 'no', '1', 't', 'true', 'on',
              'y', 'yes')
    @mock.patch('cinder.volume.volume_types.update')
    @mock.patch('cinder.volume.volume_types.get_volume_type')
    @mock.patch("cinder.api.openstack.wsgi.Request.cache_resource")
    @mock.patch("cinder.api.views.types.ViewBuilder.show")
    def test_update_with_valid_is_public_in_string(
            self, is_public, mock_show, mock_cache_resource,
            mock_get, mock_update):
        body = {"volume_type": {"is_public": is_public}}
        req = fakes.HTTPRequest.blank('/v2/%s/types/%s' % (
            fake.PROJECT_ID, DEFAULT_VOLUME_TYPE))
        req.method = 'PUT'
        ctxt = context.RequestContext(fake.USER_ID, fake.PROJECT_ID, True)
        req.environ['cinder.context'] = ctxt
        boolean_is_public = strutils.bool_from_string(is_public)

        self.controller._update(req, DEFAULT_VOLUME_TYPE, body=body)
        mock_update.assert_called_once_with(
            ctxt, DEFAULT_VOLUME_TYPE, None, None,
            is_public=boolean_is_public)

    @mock.patch('cinder.volume.volume_types.update')
    @mock.patch('cinder.volume.volume_types.get_volume_type')
    def test_update_type_with_description_having_length_zero(
            self, mock_get_volume_type, mock_type_update):

        mock_get_volume_type.return_value = \
            {'id': DEFAULT_VOLUME_TYPE, 'name': u'vol_type_1',
             'description': u''}

        type_description = ""
        body = {"volume_type": {"description": type_description}}
        req = fakes.HTTPRequest.blank('/v2/%s/types/%s' % (
            fake.PROJECT_ID, DEFAULT_VOLUME_TYPE))
        req.method = 'PUT'
        resp = self.controller._update(req, DEFAULT_VOLUME_TYPE, body=body)
        self._check_test_results(resp,
                                 {'expected_desc': '',
                                  'expected_name': 'vol_type_1'})

    def test_update_type_with_name_too_long(self):
        type_name = 'a' * 256
        body = {"volume_type": {"name": type_name,
                                "description": ""}}
        req = fakes.HTTPRequest.blank('/v2/%s/types/%s' % (
            fake.PROJECT_ID, DEFAULT_VOLUME_TYPE))
        req.method = 'PUT'
        self.assertRaises(exception.ValidationError,
                          self.controller._update, req,
                          DEFAULT_VOLUME_TYPE, body=body)

    def test_update_type_with_description_too_long(self):
        type_description = 'a' * 256
        body = {"volume_type": {"description": type_description}}
        req = fakes.HTTPRequest.blank('/v2/%s/types/%s' % (
            fake.PROJECT_ID, DEFAULT_VOLUME_TYPE))
        req.method = 'PUT'
        self.assertRaises(exception.ValidationError,
                          self.controller._update, req,
                          DEFAULT_VOLUME_TYPE, body=body)

    @mock.patch('cinder.volume.volume_types.get_volume_type')
    @mock.patch('cinder.volume.volume_types.update')
    def test_update_non_exist(self, mock_update, mock_get_volume_type):
        mock_get_volume_type.side_effect = exception.VolumeTypeNotFound(
            volume_type_id=NOT_FOUND_VOLUME_TYPE)
        body = {"volume_type": {"name": "vol_type_1_1",
                                "description": "vol_type_desc_1_1"}}
        req = fakes.HTTPRequest.blank('/v2/%s/types/%s' % (
            fake.PROJECT_ID, NOT_FOUND_VOLUME_TYPE))
        req.method = 'PUT'

        self.assertEqual(0, len(self.notifier.notifications))
        self.assertRaises(exception.VolumeTypeNotFound,
                          self.controller._update, req,
                          NOT_FOUND_VOLUME_TYPE, body=body)
        self.assertEqual(1, len(self.notifier.notifications))

    @mock.patch('cinder.volume.volume_types.get_volume_type')
    @mock.patch('cinder.volume.volume_types.update')
    def test_update_db_fail(self, mock_update, mock_get_volume_type):
        mock_update.side_effect = exception.VolumeTypeUpdateFailed(
            id=DEFAULT_VOLUME_TYPE)
        mock_get_volume_type.return_value = fake_volume_type(
            DEFAULT_VOLUME_TYPE)

        body = {"volume_type": {"name": "vol_type_1_1",
                                "description": "vol_type_desc_1_1"}}
        req = fakes.HTTPRequest.blank('/v2/%s/types/%s' % (
            fake.PROJECT_ID, DEFAULT_VOLUME_TYPE))
        req.method = 'PUT'

        self.assertEqual(0, len(self.notifier.notifications))
        self.assertRaises(webob.exc.HTTPInternalServerError,
                          self.controller._update, req,
                          DEFAULT_VOLUME_TYPE, body=body)
        self.assertEqual(1, len(self.notifier.notifications))

    def test_update_no_name_no_description(self):
        body = {"volume_type": {}}
        req = fakes.HTTPRequest.blank('/v2/%s/types/%s' % (
            fake.PROJECT_ID, DEFAULT_VOLUME_TYPE))
        req.method = 'PUT'

        self.assertRaises(webob.exc.HTTPBadRequest,
                          self.controller._update, req,
                          DEFAULT_VOLUME_TYPE, body=body)

    def test_update_empty_name(self):
        body = {"volume_type": {"name": "  ",
                                "description": "something"}}
        req = fakes.HTTPRequest.blank('/v2/%s/types/%s' % (
            fake.PROJECT_ID, DEFAULT_VOLUME_TYPE))
        req.method = 'PUT'

        self.assertRaises(webob.exc.HTTPBadRequest,
                          self.controller._update, req,
                          DEFAULT_VOLUME_TYPE, body=body)

    @mock.patch('cinder.volume.volume_types.get_volume_type')
    @mock.patch('cinder.db.volume_type_update')
    @mock.patch('cinder.quota.VolumeTypeQuotaEngine.'
                'update_quota_resource')
    def test_update_only_name(self, mock_update_quota,
                              mock_update, mock_get):
        mock_get.return_value = return_volume_types_get_volume_type_updated(
            UPDATE_NAME_ONLY_TYPE)

        ctxt = context.RequestContext(fake.USER_ID, fake.PROJECT_ID, True)
        name = "vol_type_%s" % UPDATE_NAME_ONLY_TYPE
        updated_name = "%s_%s" % (name, UPDATE_NAME_ONLY_TYPE)
        desc = "vol_type_desc_%s" % UPDATE_NAME_ONLY_TYPE
        body = {"volume_type": {"name": name}}
        req = fakes.HTTPRequest.blank('/v2/%s/types/%s' %
                                      (fake.PROJECT_ID, UPDATE_NAME_ONLY_TYPE))
        req.method = 'PUT'
        req.environ['cinder.context'] = ctxt

        self.assertEqual(0, len(self.notifier.notifications))
        res_dict = self.controller._update(req, UPDATE_NAME_ONLY_TYPE,
                                           body=body)
        self.assertEqual(1, len(self.notifier.notifications))
        mock_update_quota.assert_called_once_with(ctxt, updated_name, name)
        self._check_test_results(res_dict,
                                 {'expected_name': updated_name,
                                  'expected_desc': desc})

    @mock.patch('cinder.volume.volume_types.update')
    @mock.patch('cinder.volume.volume_types.get_volume_type')
    def test_update_only_description(self, mock_get, mock_update):
        mock_get.return_value = return_volume_types_get_volume_type_updated(
            UPDATE_DESC_ONLY_TYPE)
        name = "vol_type_%s" % UPDATE_DESC_ONLY_TYPE
        desc = "vol_type_desc_%s" % UPDATE_DESC_ONLY_TYPE
        updated_desc = "%s_%s" % (desc, UPDATE_DESC_ONLY_TYPE)
        body = {"volume_type": {"description": updated_desc}}
        req = fakes.HTTPRequest.blank('/v2/%s/types/%s' % (
            fake.PROJECT_ID, UPDATE_DESC_ONLY_TYPE))
        req.method = 'PUT'

        self.assertEqual(0, len(self.notifier.notifications))
        res_dict = self.controller._update(req, UPDATE_DESC_ONLY_TYPE,
                                           body=body)
        self.assertEqual(1, len(self.notifier.notifications))
        self._check_test_results(res_dict,
                                 {'expected_name': name,
                                  'expected_desc': updated_desc})

    @mock.patch('cinder.volume.volume_types.update')
    @mock.patch('cinder.volume.volume_types.get_volume_type')
    def test_update_only_is_public(self, mock_get, mock_update):
        is_public = False
        mock_get.return_value = return_volume_types_get_volume_type_updated(
            DEFAULT_VOLUME_TYPE, is_public=is_public)
        name = "vol_type_%s" % DEFAULT_VOLUME_TYPE
        updated_name = '%s_%s' % (name, DEFAULT_VOLUME_TYPE)
        desc = "vol_type_desc_%s" % DEFAULT_VOLUME_TYPE
        updated_desc = "%s_%s" % (desc, DEFAULT_VOLUME_TYPE)
        body = {"volume_type": {"is_public": is_public}}
        req = fakes.HTTPRequest.blank('/v2/%s/types/%s' % (
            fake.PROJECT_ID, DEFAULT_VOLUME_TYPE))
        req.method = 'PUT'

        self.assertEqual(0, len(self.notifier.notifications))
        res_dict = self.controller._update(req, DEFAULT_VOLUME_TYPE, body=body)
        self.assertEqual(1, len(self.notifier.notifications))
        self._check_test_results(res_dict,
                                 {'expected_name': updated_name,
                                  'expected_desc': updated_desc,
                                  'is_public': False})

    def test_update_invalid_is_public(self):
        body = {"volume_type": {"name": "test",
                                "description": "something",
                                "is_public": "fake"}}
        req = fakes.HTTPRequest.blank('/v2/%s/types/%s' % (
            fake.PROJECT_ID, DEFAULT_VOLUME_TYPE))
        req.method = 'PUT'

        self.assertRaises(exception.ValidationError,
                          self.controller._update, req,
                          DEFAULT_VOLUME_TYPE, body=body)

    @mock.patch('cinder.volume.volume_types.update')
    @mock.patch('cinder.volume.volume_types.get_volume_type')
    def test_rename_existing_name(self, mock_get, mock_update):
        id = UPDATE_NAME_AFTER_DELETE_TYPE
        name = "vol_type_%s" % id
        updated_name = "%s_%s" % (name, id)
        desc = "vol_type_desc_%s" % id
        mock_update.side_effect = exception.VolumeTypeExists(
            id=id, name=name)
        mock_get.return_value = return_volume_types_get_volume_type_updated(
            UPDATE_NAME_AFTER_DELETE_TYPE)
        # first attempt fail
        body = {"volume_type": {"name": name}}
        req = fakes.HTTPRequest.blank('/v2/%s/types/%s' % (
            fake.PROJECT_ID, UPDATE_NAME_AFTER_DELETE_TYPE))
        req.method = 'PUT'

        self.assertEqual(0, len(self.notifier.notifications))
        self.assertRaises(webob.exc.HTTPConflict,
                          self.controller._update, req,
                          UPDATE_NAME_AFTER_DELETE_TYPE, body=body)

        self.assertEqual(1, len(self.notifier.notifications))

        # delete
        self.notifier.reset()
        self.mock_object(volume_types, 'destroy',
                         return_volume_types_destroy)
        req = fakes.HTTPRequest.blank('/v2/%s/types/%s' % (
            fake.PROJECT_ID, UPDATE_NAME_AFTER_DELETE_TYPE))
        self.assertEqual(0, len(self.notifier.notifications))
        self.controller._delete(req, UPDATE_NAME_AFTER_DELETE_TYPE)
        self.assertEqual(1, len(self.notifier.notifications))

        # update again
        mock_update.side_effect = mock.MagicMock()
        body = {"volume_type": {"name": updated_name}}
        req = fakes.HTTPRequest.blank('/v2/%s/types/%s' % (
            fake.PROJECT_ID, UPDATE_NAME_AFTER_DELETE_TYPE))
        req.method = 'PUT'

        self.notifier.reset()
        self.assertEqual(0, len(self.notifier.notifications))
        res_dict = self.controller._update(req, UPDATE_NAME_AFTER_DELETE_TYPE,
                                           body=body)
        self._check_test_results(res_dict,
                                 {'expected_name': name,
                                  'expected_desc': desc})
        self.assertEqual(1, len(self.notifier.notifications))

    @mock.patch('cinder.volume.volume_types.update')
    @mock.patch('cinder.volume.volume_types.get_volume_type')
    @mock.patch('cinder.policy.authorize')
    def test_update_with_non_admin(self, mock_policy_authorize, mock_get,
                                   mock_update):

        # allow policy authorized user to update type
        mock_policy_authorize.return_value = None
        mock_get.return_value = return_volume_types_get_volume_type_updated(
            DEFAULT_VOLUME_TYPE, is_public=False)
        name = "vol_type_%s" % DEFAULT_VOLUME_TYPE
        updated_name = "%s_%s" % (name, DEFAULT_VOLUME_TYPE)
        desc = "vol_type_desc_%s" % DEFAULT_VOLUME_TYPE
        updated_desc = "%s_%s" % (desc, DEFAULT_VOLUME_TYPE)
        body = {"volume_type": {"name": updated_name,
                                "description": updated_desc,
                                "is_public": False}}
        req = fakes.HTTPRequest.blank('/v2/%s/types/%s' % (
            fake.PROJECT_ID, DEFAULT_VOLUME_TYPE),
            use_admin_context=False)

        req.method = 'PUT'

        self.assertEqual(0, len(self.notifier.notifications))
        res_dict = self.controller._update(req, DEFAULT_VOLUME_TYPE, body=body)
        self.assertEqual(1, len(self.notifier.notifications))
        self._check_test_results(res_dict,
                                 {'expected_desc': updated_desc,
                                  'expected_name': updated_name,
                                  'is_public': False})

        # non policy authorized user fails to update type
        mock_policy_authorize.side_effect = (
            exception.PolicyNotAuthorized(action='type_update'))
        self.assertRaises(exception.PolicyNotAuthorized,
                          self.controller._update,
                          req, DEFAULT_VOLUME_TYPE, body=body)

    def _check_test_results(self, results, expected_results):
        self.assertEqual(1, len(results))
        self.assertEqual(expected_results['expected_desc'],
                         results['volume_type']['description'])
        if expected_results.get('expected_name'):
            self.assertEqual(expected_results['expected_name'],
                             results['volume_type']['name'])
        if expected_results.get('is_public') is not None:
            self.assertEqual(expected_results['is_public'],
                             results['volume_type']['is_public'])

    def test_update_with_name_null(self):
        body = {"volume_type": {"name": None}}
        req = fakes.HTTPRequest.blank('/v2/%s/types/%s' % (
            fake.PROJECT_ID, DEFAULT_VOLUME_TYPE))
        req.method = 'PUT'

        self.assertRaises(webob.exc.HTTPBadRequest,
                          self.controller._update, req,
                          DEFAULT_VOLUME_TYPE, body=body)

    @ddt.data({"volume_type": {"name": None, "description": "description"}},
              {"volume_type": {"name": None, "is_public": True}},
              {"volume_type": {"description": "description",
                               "is_public": True}})
    def test_update_volume_type(self, body):
        req = fakes.HTTPRequest.blank('/v2/%s/types/%s' % (
            fake.PROJECT_ID, DEFAULT_VOLUME_TYPE))
        req.method = 'PUT'
        ctxt = context.RequestContext(fake.USER_ID, fake.PROJECT_ID, True)
        req.environ['cinder.context'] = ctxt
        volume_type_1 = volume_types.create(ctxt, 'volume_type')
        res = self.controller._update(req, volume_type_1.get('id'), body=body)

        expected_name = body['volume_type'].get('name')
        if expected_name is not None:
            self.assertEqual(expected_name, res['volume_type']['name'])

        expected_is_public = body['volume_type'].get('is_public')
        if expected_is_public is not None:
            self.assertEqual(expected_is_public,
                             res['volume_type']['is_public'])

        self.assertEqual(body['volume_type'].get('description'),
                         res['volume_type']['description'])



#    Copyright 2016 EMC Corporation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
from oslo_utils import timeutils
import pytz
import six

from cinder import exception
from cinder import objects
from cinder.objects import fields
from cinder.tests.unit import fake_constants as fake
from cinder.tests.unit import objects as test_objects
from cinder.tests.unit.objects.test_group import fake_group

fake_group_snapshot = {
    'id': fake.GROUP_SNAPSHOT_ID,
    'user_id': fake.USER_ID,
    'project_id': fake.PROJECT_ID,
    'name': 'fake_name',
    'description': 'fake_description',
    'status': fields.GroupSnapshotStatus.CREATING,
    'group_id': fake.GROUP_ID,
}


class TestGroupSnapshot(test_objects.BaseObjectsTestCase):

    @mock.patch('cinder.db.sqlalchemy.api.group_snapshot_get',
                return_value=fake_group_snapshot)
    def test_get_by_id(self, group_snapshot_get):
        group_snapshot = objects.GroupSnapshot.get_by_id(
            self.context,
            fake.GROUP_SNAPSHOT_ID)
        self._compare(self, fake_group_snapshot, group_snapshot)

    @mock.patch('cinder.db.group_snapshot_create',
                return_value=fake_group_snapshot)
    def test_create(self, group_snapshot_create):
        fake_group_snap = fake_group_snapshot.copy()
        del fake_group_snap['id']
        group_snapshot = objects.GroupSnapshot(context=self.context,
                                               **fake_group_snap)
        group_snapshot.create()
        self._compare(self, fake_group_snapshot, group_snapshot)

    def test_create_with_id_except_exception(self):
        group_snapshot = objects.GroupSnapshot(
            context=self.context,
            **{'id': fake.GROUP_ID})
        self.assertRaises(exception.ObjectActionError, group_snapshot.create)

    @mock.patch('cinder.db.group_snapshot_update')
    def test_save(self, group_snapshot_update):
        group_snapshot = objects.GroupSnapshot._from_db_object(
            self.context, objects.GroupSnapshot(), fake_group_snapshot)
        group_snapshot.status = 'active'
        group_snapshot.save()
        group_snapshot_update.assert_called_once_with(self.context,
                                                      group_snapshot.id,
                                                      {'status': 'active'})

    @mock.patch('cinder.db.group_update',
                return_value=fake_group)
    @mock.patch('cinder.db.group_snapshot_update')
    def test_save_with_group(self, group_snapshot_update,
                             group_snapshot_cg_update):
        group = objects.Group._from_db_object(
            self.context, objects.Group(), fake_group)
        group_snapshot = objects.GroupSnapshot._from_db_object(
            self.context, objects.GroupSnapshot(), fake_group_snapshot)
        group_snapshot.name = 'foobar'
        group_snapshot.group = group
        self.assertEqual({'name': 'foobar',
                          'group': group},
                         group_snapshot.obj_get_changes())
        self.assertRaises(exception.ObjectActionError, group_snapshot.save)

    @mock.patch('oslo_utils.timeutils.utcnow', return_value=timeutils.utcnow())
    @mock.patch('cinder.db.sqlalchemy.api.group_snapshot_destroy')
    def test_destroy(self, group_snapshot_destroy, utcnow_mock):
        group_snapshot_destroy.return_value = {
            'status': fields.GroupSnapshotStatus.DELETED,
            'deleted': True,
            'deleted_at': utcnow_mock.return_value}
        group_snapshot = objects.GroupSnapshot(context=self.context,
                                               id=fake.GROUP_SNAPSHOT_ID)
        group_snapshot.destroy()
        self.assertTrue(group_snapshot_destroy.called)
        admin_context = group_snapshot_destroy.call_args[0][0]
        self.assertTrue(admin_context.is_admin)
        self.assertTrue(group_snapshot.deleted)
        self.assertEqual(fields.GroupSnapshotStatus.DELETED,
                         group_snapshot.status)
        self.assertEqual(utcnow_mock.return_value.replace(tzinfo=pytz.UTC),
                         group_snapshot.deleted_at)

    @mock.patch('cinder.objects.group.Group.get_by_id')
    @mock.patch(
        'cinder.objects.snapshot.SnapshotList.get_all_for_group_snapshot')
    def test_obj_load_attr(self, snapshotlist_get_for_cgs,
                           group_get_by_id):
        group_snapshot = objects.GroupSnapshot._from_db_object(
            self.context, objects.GroupSnapshot(), fake_group_snapshot)
        # Test group lazy-loaded field
        group = objects.Group(
            context=self.context, id=fake.GROUP_ID)
        group_get_by_id.return_value = group
        self.assertEqual(group, group_snapshot.group)
        group_get_by_id.assert_called_once_with(
            self.context, group_snapshot.group_id)
        # Test snapshots lazy-loaded field
        snapshots_objs = [objects.Snapshot(context=self.context, id=i)
                          for i in [fake.SNAPSHOT_ID, fake.SNAPSHOT2_ID,
                                    fake.SNAPSHOT3_ID]]
        snapshots = objects.SnapshotList(context=self.context,
                                         objects=snapshots_objs)
        snapshotlist_get_for_cgs.return_value = snapshots
        self.assertEqual(snapshots, group_snapshot.snapshots)
        snapshotlist_get_for_cgs.assert_called_once_with(
            self.context, group_snapshot.id)

    @mock.patch('cinder.db.sqlalchemy.api.group_snapshot_get')
    def test_refresh(self, group_snapshot_get):
        db_group_snapshot1 = fake_group_snapshot.copy()
        db_group_snapshot2 = db_group_snapshot1.copy()
        db_group_snapshot2['description'] = 'foobar'

        # On the second group_snapshot_get, return the GroupSnapshot with an
        # updated description
        group_snapshot_get.side_effect = [db_group_snapshot1,
                                          db_group_snapshot2]
        group_snapshot = objects.GroupSnapshot.get_by_id(
            self.context, fake.GROUP_SNAPSHOT_ID)
        self._compare(self, db_group_snapshot1, group_snapshot)

        # description was updated, so a GroupSnapshot refresh should have a new
        # value for that field
        group_snapshot.refresh()
        self._compare(self, db_group_snapshot2, group_snapshot)
        if six.PY3:
            call_bool = mock.call.__bool__()
        else:
            call_bool = mock.call.__nonzero__()
        group_snapshot_get.assert_has_calls(
            [mock.call(self.context,
                       fake.GROUP_SNAPSHOT_ID),
             call_bool,
             mock.call(self.context,
                       fake.GROUP_SNAPSHOT_ID)])


class TestGroupSnapshotList(test_objects.BaseObjectsTestCase):
    @mock.patch('cinder.db.group_snapshot_get_all',
                return_value=[fake_group_snapshot])
    def test_get_all(self, group_snapshot_get_all):
        group_snapshots = objects.GroupSnapshotList.get_all(self.context)
        self.assertEqual(1, len(group_snapshots))
        TestGroupSnapshot._compare(self, fake_group_snapshot,
                                   group_snapshots[0])

    @mock.patch('cinder.db.group_snapshot_get_all_by_project',
                return_value=[fake_group_snapshot])
    def test_get_all_by_project(self, group_snapshot_get_all_by_project):
        group_snapshots = objects.GroupSnapshotList.get_all_by_project(
            self.context, self.project_id)
        self.assertEqual(1, len(group_snapshots))
        TestGroupSnapshot._compare(self, fake_group_snapshot,
                                   group_snapshots[0])

    @mock.patch('cinder.db.group_snapshot_get_all_by_group',
                return_value=[fake_group_snapshot])
    def test_get_all_by_group(self, group_snapshot_get_all_by_group):
        group_snapshots = objects.GroupSnapshotList.get_all_by_group(
            self.context, self.project_id)
        self.assertEqual(1, len(group_snapshots))
        TestGroupSnapshot._compare(self, fake_group_snapshot,
                                   group_snapshots[0])



# Copyright (c) 2018 LINBIT HA Solutions GmbH
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from oslo_config import cfg
from oslo_utils import timeutils

from cinder import exception as cinder_exception
from cinder import test
from cinder.volume import configuration as conf
from cinder.volume.drivers import linstordrv as drv

CONF = cfg.CONF

CINDER_UNKNOWN = 'unknown'
LVM = 'Lvm'
LVMTHIN = 'LvmThin'
DRIVER = 'cinder.volume.drivers.linstordrv.'

RESOURCE = {
    'name': 'CV_bc3015e6-695f-4688-91f2-1deb4321e4f0',
    'volume': {
        'device_path': '/dev/drbd1000',
    },
}

RESOURCE_LIST = {
    'resourceStates': [
        {
            'rscName': 'CV_bc3015e6-695f-4688-91f2-1deb4321e4f0',
            'nodeName': 'node_one',
            'inUse': False,
            'vlmStates': [
                {
                    'vlmNr': 0,
                    'diskState': 'Diskless',
                }
            ],
        },
        {
            'rscName': 'CV_bc3015e6-695f-4688-91f2-1deb4321e4f0',
            'nodeName': 'node_two',
            'inUse': False,
            'vlmStates': [
                {
                    'vlmNr': 0,
                    'diskState': 'UpToDate',
                }
            ],
        },
    ],
    'resources': [
        {
            'name': 'CV_bc3015e6-695f-4688-91f2-1deb4321e4f0',
            'nodeId': 0,
            'nodeName': 'node_one',
            'nodeUuid': '67939f68-2b26-41b7-b32e-a20b77664aef',
            'props': [{'key': 'PeerSlots', 'value': '7'}],
            'rscDfnUuid': '03623665-35a3-4caa-aa92-0c8badbda84a',
            'uuid': '559a229e-2b97-4d20-8f6d-87778bbe2f9e',
            'vlms': [
                {
                    'backingDisk': '/dev/vg-35/f1_00000',
                    'devicePath': '/dev/drbd1000',
                    'metaDisk': 'internal',
                    'storPoolName': 'DfltStorPool',
                    'storPoolUuid': 'd2f293f5-5d73-4447-a14b-70efe01302be',
                    'vlmDfnUuid': '0eedabe4-3c20-4eff-af74-b2ec2304ab0c',
                    'vlmMinorNr': 1000,
                    'vlmNr': 0,
                    'vlmUuid': '38e48fb8-e0af-4317-8aab-aabb46db4cf8'
                }
            ]
        },
        {
            'name': 'CV_bc3015e6-695f-4688-91f2-1deb4321e4f0',
            'nodeId': 1,
            'nodeName': 'node_two',
            'nodeUuid': '82c4c5a5-8290-481e-9e35-1c71094b0cab',
            'props': [{'key': 'PeerSlots', 'value': '7'}],
            'rscDfnUuid': '03623665-35a3-4caa-aa92-0c8badbda84a',
            'rscFlags': ['DISKLESS'],
            'uuid': '23d3d331-ad0c-43f3-975b-d1048e09dc23',
            'vlms': [
                {
                    'backingDisk': 'none',
                    'devicePath': '/dev/drbd1000',
                    'metaDisk': 'internal',
                    'storPoolName': 'DfltStorPool',
                    'storPoolUuid': '85ef7894-0682-4019-b95a-1b25e81c0cb5',
                    'vlmDfnUuid': '0eedabe4-3c20-4eff-af74-b2ec2304ab0c',
                    'vlmMinorNr': 1000,
                    'vlmNr': 0,
                    'vlmUuid': 'd25b6c91-680f-4aa6-97c3-533e4bf4e659'
                }
            ]
        }
    ]
}

RESOURCE_LIST_RESP = ['node_two', 'node_one']

SNAPSHOT_LIST_RESP = ['node_two']

RESOURCE_DFN_LIST = {
    'rscDfns': [
        {
            'rscDfnPort': 7002,
            'rscDfnProps': [{'key': u'DrbdPrimarySetOn',
                             'value': u'NODE_TWO'}],
            'rscDfnSecret': u'syxflfoMqj84cUUcsqta',
            'rscDfnUuid': u'f55f0c28-455b-458f-a05d-b5f7f16b5c22',
            'rscName': u'CV_bc3015e6-695f-4688-91f2-1deb4321e4f0',
            'vlmDfns': [
                {
                    'vlmDfnUuid': u'89f6eff2-c4cd-4586-9ab8-8e850568b93b',
                    'vlmMinor': 1001,
                    'vlmNr': 0,
                    'vlmProps': [{'key': u'DrbdCurrentGi',
                                  'value': u'2286D24524D26AA'}],
                    'vlmSize': '1044480'}
            ]
        },
    ]
}

RESOURCE_DFN_LIST_RESP = [
    {
        'rd_name': u'CV_bc3015e6-695f-4688-91f2-1deb4321e4f0',
        'rd_port': 7002,
        'rd_size': 1.0,
        'rd_uuid': u'f55f0c28-455b-458f-a05d-b5f7f16b5c22',
        'vlm_dfn_uuid': u'89f6eff2-c4cd-4586-9ab8-8e850568b93b'
    }
]

NODES_LIST = {
    'nodes': [
        {
            'connectionStatus': 2,
            'name': u'node_two',
            'netInterfaces': [
                {
                    'address': u'192.168.66.113',
                    'name': u'default',
                    'stltEncryptionType': u'PLAIN',
                    'stltPort': 3366,
                    'uuid': u'224e50c3-09a8-4cf8-b701-13663a66aecd'
                }
            ],
            'props': [{'key': u'CurStltConnName', 'value': u'default'}],
            'type': u'COMBINED',
            'uuid': u'67939f68-2b26-41b7-b32e-a20b77664aef'
        },
        {
            'connectionStatus': 2,
            'name': u'node_one',
            'netInterfaces': [
                {
                    'address': u'192.168.66.115',
                    'name': u'default',
                    'stltEncryptionType': u'PLAIN',
                    'stltPort': 3366,
                    'uuid': u'36f42ec9-9999-4ad7-a889-8d7dbb498163'
                }
            ],
            'props': [{'key': u'CurStltConnName', 'value': u'default'}],
            'type': u'COMBINED',
            'uuid': u'82c4c5a5-8290-481e-9e35-1c71094b0cab'
        }
    ]
}

NODES_RESP = [
    {
        'node_address': u'192.168.66.113',
        'node_name': u'node_two',
        'node_uuid': u'67939f68-2b26-41b7-b32e-a20b77664aef'
    },
    {
        'node_address': u'192.168.66.115',
        'node_name': u'node_one',
        'node_uuid': u'82c4c5a5-8290-481e-9e35-1c71094b0cab'
    }
]

STORAGE_POOL_DEF = {
    'storPoolDfns': [
        {
            'storPoolName': u'DfltStorPool',
            'uuid': u'f51611c6-528f-4793-a87a-866d09e6733a'
        }
    ]
}

STORAGE_POOL_DEF_RESP = [
    {
        'spd_name': u'DfltStorPool',
        'spd_uuid': u'f51611c6-528f-4793-a87a-866d09e6733a'
    }
]

STORAGE_POOL_LIST = {
    'storPools': [
        {
            'driver': u'LvmThinDriver',
            'freeSpace': {
                'freeCapacity': '36700160',
                'storPoolName': u'DfltStorPool',
                'storPoolUuid': u'd2f293f5-5d73-4447-a14b-70efe01302be',
                'totalCapacity': '36700160'
            },
            'nodeName': u'node_two',
            'nodeUuid': u'67939f68-2b26-41b7-b32e-a20b77664aef',
            'props': [{'key': u'StorDriver/LvmVg', 'value': u'vg-35'},
                      {'key': u'StorDriver/ThinPool',
                       'value': u'thinpool'}],
            'staticTraits': [{'key': u'Provisioning', 'value': u'Thin'},
                             {'key': u'SupportsSnapshots',
                              'value': u'true'}],
            'storPoolDfnUuid': u'f51611c6-528f-4793-a87a-866d09e6733a',
            'storPoolName': u'DfltStorPool',
            'storPoolUuid': u'd2f293f5-5d73-4447-a14b-70efe01302be',
            'vlms': [
                {
                    'backingDisk':
                        u'/dev/vg-35/CV_bc3015e6-695f-4688-91f2-1deb4321e4f0',
                    'devicePath': u'/dev/drbd1001',
                    'metaDisk': u'internal',
                    'storPoolName': u'DfltStorPool',
                    'storPoolUuid': u'd2f293f5-5d73-4447-a14b-70efe01302be',
                    'vlmDfnUuid': u'89f6eff2-c4cd-4586-9ab8-8e850568b93b',
                    'vlmMinorNr': 1001,
                    'vlmNr': 0,
                    'vlmUuid': u'b91392ae-904a-4bc6-862f-9c7aca629b35'
                },
                {
                    'backingDisk': u'/dev/vg-35/f1_00000',
                    'devicePath': u'/dev/drbd1000',
                    'metaDisk': u'internal',
                    'storPoolName': u'DfltStorPool',
                    'storPoolUuid': u'd2f293f5-5d73-4447-a14b-70efe01302be',
                    'vlmDfnUuid': u'0eedabe4-3c20-4eff-af74-b2ec2304ab0c',
                    'vlmMinorNr': 1000,
                    'vlmNr': 0,
                    'vlmUuid': u'38e48fb8-e0af-4317-8aab-aabb46db4cf8'
                }
            ]
        },
        {
            'driver': u'DisklessDriver',
            'freeSpace': {
                'freeCapacity': '9223372036854775807',
                'storPoolName': u'DfltStorPool',
                'storPoolUuid': u'85ef7894-0682-4019-b95a-1b25e81c0cb5',
                'totalCapacity': '9223372036854775807'
            },
            'nodeName': u'node_one',
            'nodeUuid': u'82c4c5a5-8290-481e-9e35-1c71094b0cab',
            'staticTraits': [{'key': u'SupportsSnapshots',
                              'value': u'false'}],
            'storPoolDfnUuid': u'f51611c6-528f-4793-a87a-866d09e6733a',
            'storPoolName': u'DfltStorPool',
            'storPoolUuid': u'85ef7894-0682-4019-b95a-1b25e81c0cb5',
            'vlms': [
                {
                    'backingDisk': u'none',
                    'devicePath': u'/dev/drbd1001',
                    'metaDisk': u'internal',
                    'storPoolName': u'DfltStorPool',
                    'storPoolUuid': u'85ef7894-0682-4019-b95a-1b25e81c0cb5',
                    'vlmDfnUuid': u'89f6eff2-c4cd-4586-9ab8-8e850568b93b',
                    'vlmMinorNr': 1001,
                    'vlmNr': 0,
                    'vlmUuid': u'4c63ee46-acb0-4aa5-8758-8fa8f65fdd5a'
                },
                {
                    'backingDisk': u'none',
                    'devicePath': u'/dev/drbd1000',
                    'metaDisk': u'internal',
                    'storPoolName': u'DfltStorPool',
                    'storPoolUuid': u'85ef7894-0682-4019-b95a-1b25e81c0cb5',
                    'vlmDfnUuid': u'0eedabe4-3c20-4eff-af74-b2ec2304ab0c',
                    'vlmMinorNr': 1000,
                    'vlmNr': 0,
                    'vlmUuid': u'd25b6c91-680f-4aa6-97c3-533e4bf4e659'
                }
            ]
        }
    ]
}

STORAGE_POOL_LIST_RESP = [
    {
        'driver_name': 'LvmThin',
        'node_name': u'node_two',
        'node_uuid': u'67939f68-2b26-41b7-b32e-a20b77664aef',
        'sp_cap': 35.0,
        'sp_free': 35.0,
        'sp_name': u'DfltStorPool',
        'sp_uuid': u'd2f293f5-5d73-4447-a14b-70efe01302be',
        'sp_vlms_uuid': [u'89f6eff2-c4cd-4586-9ab8-8e850568b93b',
                         u'0eedabe4-3c20-4eff-af74-b2ec2304ab0c']
    },
    {
        'driver_name': u'DisklessDriver',
        'node_name': u'node_one',
        'node_uuid': u'82c4c5a5-8290-481e-9e35-1c71094b0cab',
        'sp_cap': 0.0,
        'sp_free': -1.0,
        'sp_name': u'DfltStorPool',
        'sp_uuid': u'85ef7894-0682-4019-b95a-1b25e81c0cb5',
        'sp_vlms_uuid': [u'89f6eff2-c4cd-4586-9ab8-8e850568b93b',
                         u'0eedabe4-3c20-4eff-af74-b2ec2304ab0c']
    }
]

VOLUME_LIST = {
    'resourceStates': [
        {
            'inUse': False,
            'nodeName': u'wp-u16-cinder-dev-lg',
            'rscName': u'CV_bc3015e6-695f-4688-91f2-1deb4321e4f0',
            'vlmStates': [{'diskState': u'Diskless', 'vlmNr': 0}]
        },
        {
            'nodeName': u'wp-u16-cinder-dev-1', 'rscName': u'foo'
        },
        {
            'inUse': False,
            'nodeName': u'wp-u16-cinder-dev-1',
            'rscName': u'CV_bc3015e6-695f-4688-91f2-1deb4321e4f0',
            'vlmStates': [{'diskState': u'UpToDate', 'vlmNr': 0}]
        }
    ],
    'resources': [
        {
            'name': u'CV_bc3015e6-695f-4688-91f2-1deb4321e4f0',
            'nodeId': 0,
            'nodeName': u'wp-u16-cinder-dev-1',
            'nodeUuid': u'67939f68-2b26-41b7-b32e-a20b77664aef',
            'props': [{'key': u'PeerSlots', 'value': u'7'}],
            'rscDfnUuid': u'f55f0c28-455b-458f-a05d-b5f7f16b5c22',
            'uuid': u'2da61a7a-83b7-41d1-8a96-3a1a118dfba2',
            'vlms': [
                {
                    'backingDisk':
                        u'/dev/vg-35/CV_bc3015e6-695f-4688-91f2-' +
                        u'1deb4321e4f0_00000',
                    'devicePath': u'/dev/drbd1001',
                    'metaDisk': u'internal',
                    'storPoolName': u'DfltStorPool',
                    'storPoolUuid': u'd2f293f5-5d73-4447-a14b-70efe01302be',
                    'vlmDfnUuid': u'89f6eff2-c4cd-4586-9ab8-8e850568b93b',
                    'vlmMinorNr': 1001,
                    'vlmNr': 0,
                    'vlmUuid': u'b91392ae-904a-4bc6-862f-9c7aca629b35'
                }
            ]
        },
        {
            'name': u'CV_bc3015e6-695f-4688-91f2-1deb4321e4f0',
            'nodeId': 1,
            'nodeName': u'wp-u16-cinder-dev-lg',
            'nodeUuid': u'82c4c5a5-8290-481e-9e35-1c71094b0cab',
            'props': [{'key': u'PeerSlots', 'value': u'7'}],
            'rscDfnUuid': u'f55f0c28-455b-458f-a05d-b5f7f16b5c22',
            'rscFlags': [u'DISKLESS'],
            'uuid': u'bd6472d1-dc3c-4d41-a5f0-f44271c05680',
            'vlms': [
                {
                    'backingDisk': u'none',
                    'devicePath': u'/dev/drbd1001',
                    'metaDisk': u'internal',
                    'storPoolName': u'DfltStorPool',
                    'storPoolUuid': u'85ef7894-0682-4019-b95a-1b25e81c0cb5',
                    'vlmDfnUuid': u'89f6eff2-c4cd-4586-9ab8-8e850568b93b',
                    'vlmMinorNr': 1001,
                    'vlmNr': 0,
                    'vlmUuid': u'4c63ee46-acb0-4aa5-8758-8fa8f65fdd5a'
                }
            ]
        }
    ]
}

VOLUME_LIST_RESP = [
    {
        'node_name': u'wp-u16-cinder-dev-1',
        'rd_name': u'CV_bc3015e6-695f-4688-91f2-1deb4321e4f0',
        'volume': [
            {
                'backingDisk': u'/dev/vg-35/CV_bc3015e6-695f-4688-91f2-' +
                               u'1deb4321e4f0_00000',
                'devicePath': u'/dev/drbd1001',
                'metaDisk': u'internal',
                'storPoolName': u'DfltStorPool',
                'storPoolUuid': u'd2f293f5-5d73-4447-a14b-70efe01302be',
                'vlmDfnUuid': u'89f6eff2-c4cd-4586-9ab8-8e850568b93b',
                'vlmMinorNr': 1001,
                'vlmNr': 0,
                'vlmUuid': u'b91392ae-904a-4bc6-862f-9c7aca629b35'
            }
        ]
    },
    {
        'node_name': u'wp-u16-cinder-dev-lg',
        'rd_name': u'CV_bc3015e6-695f-4688-91f2-1deb4321e4f0',
        'volume': [
            {
                'backingDisk': u'none',
                'devicePath': u'/dev/drbd1001',
                'metaDisk': u'internal',
                'storPoolName': u'DfltStorPool',
                'storPoolUuid': u'85ef7894-0682-4019-b95a-1b25e81c0cb5',
                'vlmDfnUuid': u'89f6eff2-c4cd-4586-9ab8-8e850568b93b',
                'vlmMinorNr': 1001,
                'vlmNr': 0,
                'vlmUuid': u'4c63ee46-acb0-4aa5-8758-8fa8f65fdd5a'
            }
        ]
    }
]

VOLUME_STATS_RESP = {
    'driver_version': '0.0.7',
    'pools': [
        {
            'QoS_support': False,
            'backend_state': 'up',
            'filter_function': None,
            'free_capacity_gb': 35.0,
            'goodness_function': None,
            'location_info': 'linstor://localhost',
            'max_over_subscription_ratio': 0,
            'multiattach': False,
            'pool_name': 'lin-test-driver',
            'provisioned_capacity_gb': 1.0,
            'reserved_percentage': 0,
            'thick_provisioning_support': False,
            'thin_provisioning_support': True,
            'total_capacity_gb': 35.0,
            'total_volumes': 1,
        }
    ],
    'vendor_name': 'LINBIT',
    'volume_backend_name': 'lin-test-driver'
}

CINDER_VOLUME = {
    'id': 'bc3015e6-695f-4688-91f2-1deb4321e4f0',
    'name': 'test-lin-vol',
    'size': 1,
    'volume_type_id': 'linstor',
    'created_at': timeutils.utcnow()
}

SNAPSHOT = {
    'id': 'bc3015e6-695f-4688-91f2-1deb4321e4f0',
    'volume_id': 'bc3015e6-695f-4688-91f2-1deb4321e4f0',
    'volume_size': 1
}

VOLUME_NAMES = {
    'linstor': 'CV_bc3015e6-695f-4688-91f2-1deb4321e4f0',
    'cinder': 'bc3015e6-695f-4688-91f2-1deb4321e4f0',
    'snap': 'SN_bc3015e6-695f-4688-91f2-1deb4321e4f0',
}


class LinstorAPIFakeDriver(object):

    def fake_api_ping(self):
        return 1234

    def fake_api_resource_list(self):
        return RESOURCE_LIST

    def fake_api_node_list(self):
        return NODES_LIST

    def fake_api_storage_pool_dfn_list(self):
        return STORAGE_POOL_DEF

    def fake_api_storage_pool_list(self):
        return STORAGE_POOL_LIST

    def fake_api_volume_list(self):
        return VOLUME_LIST

    def fake_api_resource_dfn_list(self):
        return RESOURCE_DFN_LIST

    def fake_api_snapshot_list(self):
        return SNAPSHOT_LIST_RESP


class LinstorBaseDriverTestCase(test.TestCase):

    def __init__(self, *args, **kwargs):
        super(LinstorBaseDriverTestCase, self).__init__(*args, **kwargs)

    def setUp(self):
        super(LinstorBaseDriverTestCase, self).setUp()

        if drv is None:
            return

        self._mock = mock.Mock()
        self._fake_driver = LinstorAPIFakeDriver()

        self.configuration = mock.Mock(conf.Configuration)

        self.driver = drv.LinstorBaseDriver(
            configuration=self.configuration)
        self.driver.VERSION = '0.0.7'
        self.driver.default_rsc_size = 1
        self.driver.default_vg_name = 'vg-1'
        self.driver.default_downsize_factor = int('4096')
        self.driver.default_pool = STORAGE_POOL_DEF_RESP[0]['spd_name']
        self.driver.host_name = 'node_one'
        self.driver.diskless = True
        self.driver.default_uri = 'linstor://localhost'
        self.driver.default_backend_name = 'lin-test-driver'
        self.driver.configuration.reserved_percentage = 0
        self.driver.configuration.max_over_subscription_ratio = 0

    @mock.patch(DRIVER + 'LinstorBaseDriver._ping')
    def test_ping(self, m_ping):
        m_ping.return_value = self._fake_driver.fake_api_ping()

        val = self.driver._ping()
        expected = 1234
        self.assertEqual(expected, val)

    @mock.patch('uuid.uuid4')
    def test_clean_uuid(self, m_uuid):
        m_uuid.return_value = u'bd6472d1-dc3c-4d41-a5f0-f44271c05680'

        val = self.driver._clean_uuid()
        expected = u'bd6472d1-dc3c-4d41-a5f0-f44271c05680'
        self.assertEqual(expected, val)

    # Test volume size conversions
    def test_unit_conversions_to_linstor(self):
        val = self.driver._vol_size_to_linstor(1)
        expected = 1044480   # 1048575 - 4096
        self.assertEqual(expected, val)

    def test_unit_conversions_to_cinder(self):
        val = self.driver._vol_size_to_cinder(1048576)
        expected = 1
        self.assertEqual(expected, val)

    def test_is_clean_volume_name(self):
        val = self.driver._is_clean_volume_name(VOLUME_NAMES['cinder'],
                                                drv.DM_VN_PREFIX)
        expected = VOLUME_NAMES['linstor']
        self.assertEqual(expected, val)

    def test_snapshot_name_from_cinder_snapshot(self):
        val = self.driver._snapshot_name_from_cinder_snapshot(
            SNAPSHOT)
        expected = VOLUME_NAMES['snap']
        self.assertEqual(expected, val)

    def test_cinder_volume_name_from_drbd_resource(self):
        val = self.driver._cinder_volume_name_from_drbd_resource(
            VOLUME_NAMES['linstor'])
        expected = VOLUME_NAMES['cinder']
        self.assertEqual(expected, val)

    def test_drbd_resource_name_from_cinder_snapshot(self):
        val = self.driver._drbd_resource_name_from_cinder_snapshot(
            SNAPSHOT)
        expected = VOLUME_NAMES['linstor']
        self.assertEqual(expected, val)

    def test_drbd_resource_name_from_cinder_volume(self):
        val = self.driver._drbd_resource_name_from_cinder_volume(
            CINDER_VOLUME)
        expected = VOLUME_NAMES['linstor']
        self.assertEqual(expected, val)

    @mock.patch(DRIVER + 'LinstorBaseDriver._get_api_resource_list')
    def test_get_rcs_path(self, m_rsc_list):
        m_rsc_list.return_value = self._fake_driver.fake_api_resource_list()

        val = self.driver._get_rsc_path(VOLUME_NAMES['linstor'])
        expected = '/dev/drbd1000'
        self.assertEqual(expected, val)

    @mock.patch(DRIVER + 'LinstorBaseDriver._get_api_resource_list')
    def test_get_local_path(self, m_rsc_list):
        m_rsc_list.return_value = self._fake_driver.fake_api_resource_list()

        val = self.driver._get_local_path(CINDER_VOLUME)
        expected = '/dev/drbd1000'
        self.assertEqual(expected, val)

    @mock.patch(DRIVER + 'LinstorBaseDriver._get_api_storage_pool_dfn_list')
    def test_get_spd(self, m_spd_list):
        m_spd_list.return_value = (
            self._fake_driver.fake_api_storage_pool_dfn_list())

        val = self.driver._get_spd()
        expected = STORAGE_POOL_DEF_RESP
        self.assertEqual(expected, val)

    @mock.patch(DRIVER + 'LinstorBaseDriver._get_api_storage_pool_list')
    def test_get_storage_pool(self, m_sp_list):
        m_sp_list.return_value = (
            self._fake_driver.fake_api_storage_pool_list())

        val = self.driver._get_storage_pool()
        expected = STORAGE_POOL_LIST_RESP
        self.assertEqual(expected, val)

    @mock.patch(DRIVER + 'LinstorBaseDriver._get_api_resource_dfn_list')
    def test_get_resource_definitions(self, m_rscd_list):
        m_rscd_list.return_value = (
            self._fake_driver.fake_api_resource_dfn_list())

        val = self.driver._get_resource_definitions()
        expected = RESOURCE_DFN_LIST_RESP
        self.assertEqual(expected, val)

    @mock.patch(DRIVER + 'LinstorBaseDriver._get_snapshot_nodes')
    def test_get_snapshot_nodes(self, m_rsc_list):
        m_rsc_list.return_value = self._fake_driver.fake_api_snapshot_list()

        val = self.driver._get_snapshot_nodes(VOLUME_NAMES['linstor'])
        expected = SNAPSHOT_LIST_RESP
        self.assertEqual(expected, val)

    @mock.patch(DRIVER + 'LinstorBaseDriver._get_api_nodes_list')
    def test_get_linstor_nodes(self, m_node_list):
        m_node_list.return_value = self._fake_driver.fake_api_node_list()

        val = self.driver._get_linstor_nodes()
        expected = RESOURCE_LIST_RESP
        self.assertEqual(expected, val)

    @mock.patch(DRIVER + 'LinstorBaseDriver._get_api_nodes_list')
    def test_get_nodes(self, m_node_list):
        m_node_list.return_value = self._fake_driver.fake_api_node_list()

        val = self.driver._get_nodes()
        expected = NODES_RESP
        self.assertEqual(expected, val)

    @mock.patch(DRIVER + 'LinstorBaseDriver.get_goodness_function')
    @mock.patch(DRIVER + 'LinstorBaseDriver.get_filter_function')
    @mock.patch(DRIVER + 'LinstorBaseDriver._get_api_resource_dfn_list')
    @mock.patch(DRIVER + 'LinstorBaseDriver._get_api_storage_pool_list')
    def test_get_volume_stats(self,
                              m_sp_list,
                              m_rscd_list,
                              m_filter,
                              m_goodness):
        m_sp_list.return_value = (
            self._fake_driver.fake_api_storage_pool_list())
        m_rscd_list.return_value = (
            self._fake_driver.fake_api_resource_dfn_list())
        m_filter.return_value = None
        m_goodness.return_value = None

        val = self.driver._get_volume_stats()
        expected = VOLUME_STATS_RESP
        self.assertEqual(expected, val)

    @mock.patch(DRIVER + 'LinstorBaseDriver._get_api_resource_list')
    @mock.patch(DRIVER + 'LinstorBaseDriver._check_api_reply')
    @mock.patch(DRIVER + 'LinstorBaseDriver._api_snapshot_create')
    def test_create_snapshot_fail(self,
                                  m_snap_create,
                                  m_api_reply,
                                  m_rsc_list):
        m_snap_create.return_value = None
        m_rsc_list.return_value = self._fake_driver.fake_api_resource_list()
        m_api_reply.return_value = False

        self.assertRaises(cinder_exception.VolumeBackendAPIException,
                          self.driver.create_snapshot, SNAPSHOT)

    @mock.patch(DRIVER + 'LinstorBaseDriver._get_api_resource_list')
    @mock.patch(DRIVER + 'LinstorBaseDriver._check_api_reply')
    @mock.patch(DRIVER + 'LinstorBaseDriver._api_snapshot_create')
    def test_create_snapshot_success(self,
                                     m_snap_create,
                                     m_api_reply,
                                     m_rsc_list):
        m_snap_create.return_value = None
        m_rsc_list.return_value = self._fake_driver.fake_api_resource_list()
        m_api_reply.return_value = True

        # No exception should be raised
        self.assertIsNone(self.driver.create_snapshot(SNAPSHOT))

    @mock.patch(DRIVER + 'LinstorBaseDriver._get_api_resource_dfn_list')
    @mock.patch(DRIVER + 'LinstorBaseDriver._get_api_resource_list')
    @mock.patch(DRIVER + 'LinstorBaseDriver._check_api_reply')
    @mock.patch(DRIVER + 'LinstorBaseDriver._api_snapshot_delete')
    def test_delete_snapshot_fail(self,
                                  m_snap_delete,
                                  m_api_reply,
                                  m_rsc_list,
                                  m_rsc_dfn_list):
        m_snap_delete.return_value = None
        m_api_reply.return_value = False
        m_rsc_list.return_value = self._fake_driver.fake_api_resource_list()
        m_rsc_dfn_list.return_value = (
            self._fake_driver.fake_api_resource_dfn_list())

        self.assertRaises(cinder_exception.VolumeBackendAPIException,
                          self.driver.delete_snapshot, SNAPSHOT)

    @mock.patch(DRIVER + 'LinstorBaseDriver._api_rsc_dfn_delete')
    @mock.patch(DRIVER + 'LinstorBaseDriver._get_api_resource_list')
    @mock.patch(DRIVER + 'LinstorBaseDriver._check_api_reply')
    @mock.patch(DRIVER + 'LinstorBaseDriver._api_snapshot_delete')
    def test_delete_snapshot_success(self,
                                     m_snap_delete,
                                     m_api_reply,
                                     m_rsc_list,
                                     m_rsc_dfn_delete):
        m_snap_delete.return_value = None
        m_api_reply.return_value = True
        m_rsc_list.return_value = self._fake_driver.fake_api_resource_list()
        m_rsc_dfn_delete.return_value = True

        # No exception should be raised
        self.driver.delete_snapshot(SNAPSHOT)

    @mock.patch(DRIVER + 'LinstorBaseDriver._get_api_storage_pool_list')
    @mock.patch(DRIVER + 'LinstorBaseDriver._api_volume_dfn_set_sp')
    @mock.patch(DRIVER + 'LinstorBaseDriver._get_api_volume_extend')
    @mock.patch(DRIVER + 'LinstorBaseDriver._api_rsc_create')
    @mock.patch(DRIVER + 'LinstorBaseDriver._api_snapshot_resource_restore')
    @mock.patch(DRIVER + 'LinstorBaseDriver._get_linstor_nodes')
    @mock.patch(DRIVER + 'LinstorBaseDriver._api_snapshot_volume_dfn_restore')
    @mock.patch(DRIVER + 'LinstorBaseDriver._check_api_reply')
    @mock.patch(DRIVER + 'LinstorBaseDriver._api_rsc_dfn_create')
    def test_create_volume_from_snapshot(self,
                                         m_rsc_dfn_create,
                                         m_api_reply,
                                         m_snap_vd_restore,
                                         m_lin_nodes,
                                         m_snap_rsc_restore,
                                         m_rsc_create,
                                         m_vol_extend,
                                         m_vol_dfn,
                                         m_sp_list):
        m_rsc_dfn_create.return_value = True
        m_api_reply.return_value = True
        m_snap_vd_restore.return_value = True
        m_nodes = []
        m_nodes.append('for test')
        m_nodes.remove('for test')
        m_lin_nodes.return_value = m_nodes
        m_snap_rsc_restore.return_value = True
        m_rsc_create.return_value = True
        m_vol_extend.return_value = True
        m_vol_dfn.return_value = True
        m_sp_list.return_value = (
            self._fake_driver.fake_api_storage_pool_list())

        self.assertIsNone(self.driver.create_volume_from_snapshot(
            CINDER_VOLUME, SNAPSHOT))

    @mock.patch(DRIVER + 'LinstorBaseDriver._check_api_reply')
    @mock.patch(DRIVER + 'LinstorBaseDriver._api_rsc_create')
    @mock.patch(DRIVER + 'LinstorBaseDriver._api_volume_dfn_create')
    @mock.patch(DRIVER + 'LinstorBaseDriver._api_rsc_dfn_create')
    @mock.patch(DRIVER + 'LinstorBaseDriver._api_storage_pool_create')
    @mock.patch(DRIVER + 'LinstorBaseDriver._get_api_storage_pool_dfn_list')
    @mock.patch(DRIVER + 'LinstorBaseDriver._get_api_nodes_list')
    @mock.patch(DRIVER + 'LinstorBaseDriver._get_api_storage_pool_list')
    def test_create_volume_fail_no_linstor_nodes(self,
                                                 m_sp_list,
                                                 m_node_list,
                                                 m_spd_list,
                                                 m_sp_create,
                                                 m_rsc_dfn_create,
                                                 m_vol_dfn_create,
                                                 m_rsc_create,
                                                 m_api_reply):
        m_sp_list.return_value = []
        m_node_list.return_value = []
        m_spd_list.return_value = (
            self._fake_driver.fake_api_storage_pool_dfn_list())
        m_sp_create.return_value = True
        m_rsc_dfn_create.return_value = True
        m_vol_dfn_create.return_value = True
        m_rsc_create.return_value = True
        m_api_reply.return_value = True

        test_volume = CINDER_VOLUME
        test_volume['migration_status'] = ('migrating:',
                                           str(VOLUME_NAMES['cinder']))
        test_volume['display_name'] = 'test_volume'
        test_volume['host'] = 'node_one'
        test_volume['size'] = 1

        self.assertRaises(cinder_exception.VolumeBackendAPIException,
                          self.driver.create_volume, test_volume)

    @mock.patch(DRIVER + 'LinstorBaseDriver._check_api_reply')
    @mock.patch(DRIVER + 'LinstorBaseDriver._api_rsc_create')
    @mock.patch(DRIVER + 'LinstorBaseDriver._api_volume_dfn_create')
    @mock.patch(DRIVER + 'LinstorBaseDriver._api_rsc_dfn_create')
    @mock.patch(DRIVER + 'LinstorBaseDriver._api_storage_pool_create')
    @mock.patch(DRIVER + 'LinstorBaseDriver._get_api_storage_pool_dfn_list')
    @mock.patch(DRIVER + 'LinstorBaseDriver._get_api_nodes_list')
    @mock.patch(DRIVER + 'LinstorBaseDriver._get_api_storage_pool_list')
    def test_create_volume_fail_rsc_create(self,
                                           m_sp_list,
                                           m_node_list,
                                           m_spd_list,
                                           m_sp_create,
                                           m_rsc_dfn_create,
                                           m_vol_dfn_create,
                                           m_rsc_create,
                                           m_api_reply):
        m_sp_list.return_value = (
            self._fake_driver.fake_api_storage_pool_list())
        m_node_list.return_value = self._fake_driver.fake_api_node_list()
        m_spd_list.return_value = (
            self._fake_driver.fake_api_storage_pool_dfn_list())
        m_sp_create.return_value = True
        m_rsc_dfn_create.return_value = True
        m_vol_dfn_create.return_value = True
        m_rsc_create.return_value = True
        m_api_reply.return_value = False

        test_volume = CINDER_VOLUME
        test_volume['migration_status'] = ('migrating:',
                                           str(VOLUME_NAMES['cinder']))
        test_volume['display_name'] = 'test_volume'
        test_volume['host'] = 'node_one'
        test_volume['size'] = 1

        self.assertRaises(cinder_exception.VolumeBackendAPIException,
                          self.driver.create_volume, test_volume)

    @mock.patch(DRIVER + 'LinstorBaseDriver._check_api_reply')
    @mock.patch(DRIVER + 'LinstorBaseDriver._api_rsc_create')
    @mock.patch(DRIVER + 'LinstorBaseDriver._api_volume_dfn_create')
    @mock.patch(DRIVER + 'LinstorBaseDriver._api_rsc_dfn_create')
    @mock.patch(DRIVER + 'LinstorBaseDriver._api_storage_pool_create')
    @mock.patch(DRIVER + 'LinstorBaseDriver._get_api_storage_pool_dfn_list')
    @mock.patch(DRIVER + 'LinstorBaseDriver._get_api_nodes_list')
    @mock.patch(DRIVER + 'LinstorBaseDriver._get_api_storage_pool_list')
    def test_create_volume(self,
                           m_sp_list,
                           m_node_list,
                           m_spd_list,
                           m_sp_create,
                           m_rsc_dfn_create,
                           m_vol_dfn_create,
                           m_rsc_create,
                           m_api_reply):
        m_sp_list.return_value = (
            self._fake_driver.fake_api_storage_pool_list())
        m_node_list.return_value = self._fake_driver.fake_api_node_list()
        m_spd_list.return_value = (
            self._fake_driver.fake_api_storage_pool_dfn_list())
        m_sp_create.return_value = True
        m_rsc_dfn_create.return_value = True
        m_vol_dfn_create.return_value = True
        m_rsc_create.return_value = True
        m_api_reply.return_value = True

        test_volume = CINDER_VOLUME
        test_volume['migration_status'] = ('migrating:',
                                           str(VOLUME_NAMES['cinder']))
        test_volume['display_name'] = 'test_volume'
        test_volume['host'] = 'node_one'
        test_volume['size'] = 1

        val = self.driver.create_volume(test_volume)
        expected = {}
        self.assertEqual(val, expected)

    @mock.patch(DRIVER + 'LinstorBaseDriver._check_api_reply')
    @mock.patch(DRIVER + 'LinstorBaseDriver._api_rsc_dfn_delete')
    @mock.patch(DRIVER + 'LinstorBaseDriver._api_volume_dfn_delete')
    @mock.patch(DRIVER + 'LinstorBaseDriver._api_rsc_delete')
    @mock.patch(DRIVER + 'LinstorBaseDriver._get_api_resource_list')
    def test_delete_volume_fail_incomplete(self,
                                           m_rsc_list,
                                           m_rsc_delete,
                                           m_vol_dfn_delete,
                                           m_rsc_dfn_delete,
                                           m_api_reply):
        m_rsc_list.return_value = self._fake_driver.fake_api_resource_list()
        m_rsc_delete.return_value = True
        m_vol_dfn_delete.return_value = True
        m_rsc_dfn_delete.return_value = True
        m_api_reply.return_value = False

        test_volume = CINDER_VOLUME
        test_volume['display_name'] = 'linstor_test'
        test_volume['host'] = 'node_one'
        test_volume['size'] = 1

        self.assertRaises(cinder_exception.VolumeBackendAPIException,
                          self.driver.delete_volume, test_volume)

    @mock.patch(DRIVER + 'LinstorBaseDriver._check_api_reply')
    @mock.patch(DRIVER + 'LinstorBaseDriver._api_rsc_dfn_delete')
    @mock.patch(DRIVER + 'LinstorBaseDriver._api_volume_dfn_delete')
    @mock.patch(DRIVER + 'LinstorBaseDriver._api_rsc_delete')
    @mock.patch(DRIVER + 'LinstorBaseDriver._get_api_resource_list')
    def test_delete_volume(self,
                           m_rsc_list,
                           m_rsc_delete,
                           m_vol_dfn_delete,
                           m_rsc_dfn_delete,
                           m_api_reply):
        m_rsc_list.return_value = self._fake_driver.fake_api_resource_list()
        m_rsc_delete.return_value = True
        m_vol_dfn_delete.return_value = True
        m_rsc_dfn_delete.return_value = True
        m_api_reply.return_value = True

        test_volume = CINDER_VOLUME
        test_volume['display_name'] = 'linstor_test'
        test_volume['host'] = 'node_one'
        test_volume['size'] = 1

        val = self.driver.delete_volume(test_volume)
        expected = True
        self.assertEqual(val, expected)

    @mock.patch(DRIVER + 'LinstorBaseDriver._check_api_reply')
    @mock.patch(DRIVER + 'LinstorBaseDriver._get_api_volume_extend')
    def test_extend_volume_success(self, m_vol_extend, m_api_reply):
        m_vol_extend.return_value = True
        m_api_reply.return_value = True

        # No exception should be raised
        self.driver.extend_volume(CINDER_VOLUME, 2)

    @mock.patch(DRIVER + 'LinstorBaseDriver._check_api_reply')
    @mock.patch(DRIVER + 'LinstorBaseDriver._get_api_volume_extend')
    def test_extend_volume_fail(self, m_vol_extend, m_api_reply):
        m_vol_extend.return_value = False
        m_api_reply.return_value = False

        self.assertRaises(cinder_exception.VolumeBackendAPIException,
                          self.driver.extend_volume,
                          CINDER_VOLUME,
                          2)

    def test_migrate_volume(self):
        m_ctxt = {}
        m_volume = {}
        m_host = ''

        val = self.driver.migrate_volume(m_ctxt, m_volume, m_host)
        expected = (False, None)
        self.assertEqual(val, expected)


class LinstorIscsiDriverTestCase(test.TestCase):

    def __init__(self, *args, **kwargs):
        super(LinstorIscsiDriverTestCase, self).__init__(*args, **kwargs)

    def setUp(self):
        super(LinstorIscsiDriverTestCase, self).setUp()

        self._mock = mock.Mock()
        self._fake_driver = LinstorAPIFakeDriver()

        self.configuration = mock.Mock(conf.Configuration)
        self.configuration.iscsi_helper = 'tgtadm'
        self.driver = drv.LinstorIscsiDriver(
            configuration=self.configuration, h_name='tgtadm')

        self.driver.VERSION = '0.0.7'
        self.driver.default_rsc_size = 1
        self.driver.default_vg_name = 'vg-1'
        self.driver.default_downsize_factor = int('4096')
        self.driver.default_pool = STORAGE_POOL_DEF_RESP[0]['spd_name']
        self.driver.host_name = 'node_one'
        self.driver.diskless = True
        self.driver.location_info = 'LinstorIscsi:linstor://localhost'
        self.driver.default_backend_name = 'lin-test-driver'
        self.driver.configuration.reserved_percentage = int('0')
        self.driver.configuration.max_over_subscription_ratio = int('0')

    @mock.patch(DRIVER + 'LinstorIscsiDriver._get_volume_stats')
    def test_iscsi_get_volume_stats(self, m_vol_stats):

        m_vol_stats.return_value = VOLUME_STATS_RESP

        val = self.driver.get_volume_stats()

        expected = VOLUME_STATS_RESP
        expected["storage_protocol"] = 'iSCSI'
        self.assertEqual(val, expected)

    @mock.patch(DRIVER + 'proto')
    @mock.patch(DRIVER + 'linstor')
    def test_iscsi_check_for_setup_error_pass(self, m_linstor, m_proto):
        m_linstor.return_value = True
        m_proto.return_value = True

        # No exception should be raised
        self.driver.check_for_setup_error()


class LinstorDrbdDriverTestCase(test.TestCase):

    def __init__(self, *args, **kwargs):
        super(LinstorDrbdDriverTestCase, self).__init__(*args, **kwargs)

    def setUp(self):
        super(LinstorDrbdDriverTestCase, self).setUp()

        self._mock = mock.Mock()
        self._fake_driver = LinstorAPIFakeDriver()

        self.configuration = mock.Mock(conf.Configuration)
        self.driver = drv.LinstorDrbdDriver(
            configuration=self.configuration)

        self.driver.VERSION = '0.0.7'
        self.driver.default_rsc_size = 1
        self.driver.default_vg_name = 'vg-1'
        self.driver.default_downsize_factor = int('4096')
        self.driver.default_pool = STORAGE_POOL_DEF_RESP[0]['spd_name']
        self.driver.host_name = 'node_one'
        self.driver.diskless = True
        self.driver.location_info = 'LinstorDrbd:linstor://localhost'
        self.driver.default_backend_name = 'lin-test-driver'
        self.driver.configuration.reserved_percentage = int('0')
        self.driver.configuration.max_over_subscription_ratio = int('0')

    @mock.patch(DRIVER + 'LinstorDrbdDriver._get_rsc_path')
    def test_drbd_return_drbd_config(self, m_rsc_path):
        m_rsc_path.return_value = '/dev/drbd1000'

        val = self.driver._return_drbd_config(CINDER_VOLUME)

        expected = {
            'driver_volume_type': 'local',
            'data': {
                "device_path": str(m_rsc_path.return_value)
            }
        }
        self.assertEqual(val, expected)

    @mock.patch(DRIVER + 'LinstorDrbdDriver._get_api_storage_pool_list')
    def test_drbd_node_in_sp(self, m_sp_list):
        m_sp_list.return_value = (
            self._fake_driver.fake_api_storage_pool_list())

        val = self.driver._node_in_sp('node_two')
        self.assertTrue(val)

    @mock.patch(DRIVER + 'LinstorDrbdDriver._get_volume_stats')
    def test_drbd_get_volume_stats(self, m_vol_stats):
        m_vol_stats.return_value = VOLUME_STATS_RESP

        val = self.driver.get_volume_stats()
        expected = VOLUME_STATS_RESP
        expected["storage_protocol"] = 'DRBD'
        self.assertEqual(val, expected)

    @mock.patch(DRIVER + 'proto')
    @mock.patch(DRIVER + 'linstor')
    def test_drbd_check_for_setup_error_pass(self, m_linstor, m_proto):
        m_linstor.return_value = True
        m_proto.return_value = True

        # No exception should be raised
        self.driver.check_for_setup_error()

    @mock.patch(DRIVER + 'LinstorDrbdDriver._get_rsc_path')
    @mock.patch(DRIVER + 'LinstorDrbdDriver._check_api_reply')
    @mock.patch(DRIVER + 'LinstorDrbdDriver._api_rsc_create')
    @mock.patch(DRIVER + 'LinstorDrbdDriver._node_in_sp')
    def test_drbd_initialize_connection_pass(self,
                                             m_node_sp,
                                             m_rsc_create,
                                             m_check,
                                             m_rsc_path):
        m_node_sp.return_value = True
        m_rsc_create.return_value = True
        m_check.return_value = True
        m_rsc_path.return_value = '/dev/drbd1000'

        connector = {}
        connector["host"] = 'wp-u16-cinder-dev-lg'

        val = self.driver.initialize_connection(CINDER_VOLUME, connector)

        expected = {
            'driver_volume_type': 'local',
            'data': {
                "device_path": str(m_rsc_path.return_value)
            }
        }
        self.assertEqual(val, expected)

    @mock.patch(DRIVER + 'LinstorDrbdDriver._check_api_reply')
    @mock.patch(DRIVER + 'LinstorDrbdDriver._api_rsc_delete')
    @mock.patch(DRIVER + 'LinstorDrbdDriver._node_in_sp')
    def test_drbd_terminate_connection_pass(self,
                                            m_node_sp,
                                            m_rsc_create,
                                            m_check):
        m_node_sp.return_value = True
        m_rsc_create.return_value = True
        m_check.return_value = True

        connector = {}
        connector["host"] = 'wp-u16-cinder-dev-lg'

        # No exception should be raised
        self.driver.terminate_connection(CINDER_VOLUME, connector)



# Copyright 2013 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import uuid

from oslo_serialization import jsonutils
from oslo_utils import timeutils
import webob

from cinder import context
from cinder import objects
from cinder.objects import fields
from cinder import test
from cinder.tests.unit.api import fakes
from cinder.tests.unit import fake_constants as fake
from cinder.tests.unit import fake_volume
from cinder import volume


def fake_db_volume_get(*args, **kwargs):
    return {
        'id': fake.VOLUME_ID,
        'host': 'host001',
        'status': 'available',
        'size': 5,
        'availability_zone': 'somewhere',
        'created_at': timeutils.utcnow(),
        'attach_status': fields.VolumeAttachStatus.DETACHED,
        'display_name': 'anothervolume',
        'display_description': 'Just another volume!',
        'volume_type_id': None,
        'snapshot_id': None,
        'project_id': fake.PROJECT_ID,
        'migration_status': 'migrating',
        '_name_id': fake.VOLUME2_ID,
    }


def fake_volume_api_get(*args, **kwargs):
    ctx = context.RequestContext(fake.USER_ID, fake.PROJECT_ID, True)
    db_volume = fake_db_volume_get()
    return fake_volume.fake_volume_obj(ctx, **db_volume)


def fake_volume_get_all(*args, **kwargs):
    return objects.VolumeList(objects=[fake_volume_api_get()])


def app():
    # no auth, just let environ['cinder.context'] pass through
    api = fakes.router.APIRouter()
    mapper = fakes.urlmap.URLMap()
    mapper['/v2'] = api
    return mapper


class VolumeMigStatusAttributeTest(test.TestCase):

    def setUp(self):
        super(VolumeMigStatusAttributeTest, self).setUp()
        self.mock_object(volume.api.API, 'get', fake_volume_api_get)
        self.mock_object(volume.api.API, 'get_all', fake_volume_get_all)
        self.UUID = uuid.uuid4()

    def test_get_volume_allowed(self):
        ctx = context.RequestContext(fake.USER_ID, fake.PROJECT_ID, True)
        req = webob.Request.blank('/v2/%s/volumes/%s' % (
            fake.PROJECT_ID, self.UUID))
        req.method = 'GET'
        req.environ['cinder.context'] = ctx
        res = req.get_response(app())
        vol = jsonutils.loads(res.body)['volume']
        self.assertEqual('migrating',
                         vol['os-vol-mig-status-attr:migstat'])
        self.assertEqual(fake.VOLUME2_ID,
                         vol['os-vol-mig-status-attr:name_id'])

    def test_get_volume_unallowed(self):
        ctx = context.RequestContext(fake.USER_ID, fake.PROJECT_ID, False)
        req = webob.Request.blank('/v2/%s/volumes/%s' % (
            fake.PROJECT_ID, self.UUID))
        req.method = 'GET'
        req.environ['cinder.context'] = ctx
        res = req.get_response(app())
        vol = jsonutils.loads(res.body)['volume']
        self.assertNotIn('os-vol-mig-status-attr:migstat', vol)
        self.assertNotIn('os-vol-mig-status-attr:name_id', vol)

    def test_list_detail_volumes_allowed(self):
        ctx = context.RequestContext(fake.USER_ID, fake.PROJECT_ID, True)
        req = webob.Request.blank('/v2/%s/volumes/detail' % fake.PROJECT_ID)
        req.method = 'GET'
        req.environ['cinder.context'] = ctx
        res = req.get_response(app())
        vol = jsonutils.loads(res.body)['volumes']
        self.assertEqual('migrating',
                         vol[0]['os-vol-mig-status-attr:migstat'])
        self.assertEqual(fake.VOLUME2_ID,
                         vol[0]['os-vol-mig-status-attr:name_id'])

    def test_list_detail_volumes_unallowed(self):
        ctx = context.RequestContext(fake.USER_ID, fake.PROJECT_ID, False)
        req = webob.Request.blank('/v2/%s/volumes/detail' % fake.PROJECT_ID)
        req.method = 'GET'
        req.environ['cinder.context'] = ctx
        res = req.get_response(app())
        vol = jsonutils.loads(res.body)['volumes']
        self.assertNotIn('os-vol-mig-status-attr:migstat', vol[0])
        self.assertNotIn('os-vol-mig-status-attr:name_id', vol[0])

    def test_list_simple_volumes_no_migration_status(self):
        ctx = context.RequestContext(fake.USER_ID, fake.PROJECT_ID, True)
        req = webob.Request.blank('/v2/%s/volumes' % fake.PROJECT_ID)
        req.method = 'GET'
        req.environ['cinder.context'] = ctx
        res = req.get_response(app())
        vol = jsonutils.loads(res.body)['volumes']
        self.assertNotIn('os-vol-mig-status-attr:migstat', vol[0])
        self.assertNotIn('os-vol-mig-status-attr:name_id', vol[0])



# Copyright (c) 2013 - 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from six.moves import urllib

from cinder import context
from cinder import exception
from cinder.tests.unit import fake_constants as fake
from cinder.tests.unit import fake_volume
from cinder.tests.unit.volume.drivers.dell_emc import vxflexos
from cinder.tests.unit.volume.drivers.dell_emc.vxflexos import mocks
from cinder.volume import configuration


class TestDeleteVolume(vxflexos.TestVxFlexOSDriver):
    """Test cases for ``VxFlexOSDriver.delete_volume()``"""
    def setUp(self):
        """Setup a test case environment.

        Creates a fake volume object and sets up the required API responses.
        """
        super(TestDeleteVolume, self).setUp()
        ctx = context.RequestContext('fake', 'fake', auth_token=True)

        self.volume = fake_volume.fake_volume_obj(
            ctx, **{'provider_id': fake.PROVIDER_ID})

        self.volume_name_2x_enc = urllib.parse.quote(
            urllib.parse.quote(self.driver._id_to_base64(self.volume.id))
        )

        self.HTTPS_MOCK_RESPONSES = {
            self.RESPONSE_MODE.Valid: {
                'types/Volume/instances/getByName::' +
                self.volume_name_2x_enc: self.volume.id,
                'instances/Volume::{}/action/removeMappedSdc'.format(
                    self.volume.provider_id): self.volume.provider_id,
                'instances/Volume::{}/action/removeVolume'.format(
                    self.volume.provider_id
                ): self.volume.provider_id,
            },
            self.RESPONSE_MODE.BadStatus: {
                'types/Volume/instances/getByName::' +
                self.volume_name_2x_enc: mocks.MockHTTPSResponse(
                    {
                        'errorCode': 401,
                        'message': 'BadStatus Volume Test',
                    }, 401
                ),
                'instances/Volume::{}/action/removeVolume'.format(
                    self.volume.provider_id
                ): mocks.MockHTTPSResponse(
                    {
                        'errorCode': 401,
                        'message': 'BadStatus Volume Test',
                    }, 401
                ),
            },
        }

    def test_bad_login_and_volume(self):
        self.set_https_response_mode(self.RESPONSE_MODE.BadStatus)
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.driver.delete_volume,
                          self.volume)

    def test_delete_volume(self):
        """Setting the unmap volume before delete flag for tests """
        self.override_config('vxflexos_unmap_volume_before_deletion', True,
                             configuration.SHARED_CONF_GROUP)
        self.driver.delete_volume(self.volume)



# Copyright (c) 2014, 2015, Oracle and/or its affiliates. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
ZFS Storage Appliance REST API Client Programmatic Interface
"""

import json
import ssl
import time

from oslo_log import log
from oslo_utils import strutils
import six
from six.moves import http_client
from six.moves import urllib

LOG = log.getLogger(__name__)


class Status(object):
    """Result HTTP Status"""

    def __init__(self):
        pass

    #: Request return OK
    OK = http_client.OK

    #: New resource created successfully
    CREATED = http_client.CREATED

    #: Command accepted
    ACCEPTED = http_client.ACCEPTED

    #: Command returned OK but no data will be returned
    NO_CONTENT = http_client.NO_CONTENT

    #: Bad Request
    BAD_REQUEST = http_client.BAD_REQUEST

    #: User is not authorized
    UNAUTHORIZED = http_client.UNAUTHORIZED

    #: The request is not allowed
    FORBIDDEN = http_client.FORBIDDEN

    #: The requested resource was not found
    NOT_FOUND = http_client.NOT_FOUND

    #: The request is not allowed
    NOT_ALLOWED = http_client.METHOD_NOT_ALLOWED

    #: Request timed out
    TIMEOUT = http_client.REQUEST_TIMEOUT

    #: Invalid request
    CONFLICT = http_client.CONFLICT

    #: Service Unavailable
    BUSY = http_client.SERVICE_UNAVAILABLE


class RestResult(object):
    """Result from a REST API operation"""
    def __init__(self, response=None, err=None):
        """Initialize a RestResult containing the results from a REST call.

        :param response: HTTP response
        """
        self.response = response
        self.error = err
        self.data = ""
        self.status = 0
        if self.response:
            self.status = self.response.getcode()
            result = self.response.read()
            while result:
                self.data += result.decode("utf-8")
                result = self.response.read()

        if self.error:
            self.status = self.error.code
            self.data = http_client.responses[self.status]

        LOG.debug('Response code: %s', self.status)
        LOG.debug('Response data: %s', self.data)

    def get_header(self, name):
        """Get an HTTP header with the given name from the results

        :param name: HTTP header name
        :return: The header value or None if no value is found
        """
        if self.response is None:
            return None
        info = self.response.info()
        return info.get(name)


class RestClientError(Exception):
    """Exception for ZFS REST API client errors"""
    def __init__(self, status, name="ERR_INTERNAL", message=None):

        """Create a REST Response exception

        :param status: HTTP response status
        :param name: The name of the REST API error type
        :param message: Descriptive error message returned from REST call
        """
        super(RestClientError, self).__init__(message)
        self.code = status
        self.name = name
        self.msg = message
        if status in http_client.responses:
            self.msg = http_client.responses[status]

    def __str__(self):
        return "%d %s %s" % (self.code, self.name, self.msg)


class RestClientURL(object):
    """ZFSSA urllib client"""
    def __init__(self, url, **kwargs):
        """Initialize a REST client.

        :param url: The ZFSSA REST API URL
        :key session: HTTP Cookie value of x-auth-session obtained from a
                      normal BUI login.
        :key timeout: Time in seconds to wait for command to complete.
                      (Default is 60 seconds)
        """
        self.url = url
        self.local = kwargs.get("local", False)
        self.base_path = kwargs.get("base_path", "/api")
        self.timeout = kwargs.get("timeout", 60)
        self.headers = None
        if kwargs.get('session'):
            self.headers['x-auth-session'] = kwargs.get('session')

        self.headers = {"content-type": "application/json"}
        self.do_logout = False
        self.auth_str = None

    def _path(self, path, base_path=None):
        """build rest url path"""
        if path.startswith("http://") or path.startswith("https://"):
            return path
        if base_path is None:
            base_path = self.base_path
        if not path.startswith(base_path) and not (
                self.local and ("/api" + path).startswith(base_path)):
            path = "%s%s" % (base_path, path)
        if self.local and path.startswith("/api"):
            path = path[4:]
        return self.url + path

    def _authorize(self):
        """Performs authorization setting x-auth-session"""
        self.headers['authorization'] = 'Basic %s' % self.auth_str
        if 'x-auth-session' in self.headers:
            del self.headers['x-auth-session']

        try:
            result = self.post("/access/v1")
            del self.headers['authorization']
            if result.status == http_client.CREATED:
                self.headers['x-auth-session'] = \
                    result.get_header('x-auth-session')
                self.do_logout = True
                LOG.info('ZFSSA version: %s',
                         result.get_header('x-zfssa-version'))

            elif result.status == http_client.NOT_FOUND:
                raise RestClientError(result.status, name="ERR_RESTError",
                                      message="REST Not Available: \
                                      Please Upgrade")

        except RestClientError:
            del self.headers['authorization']
            raise

    def login(self, auth_str):
        """Login to an appliance using a user name and password.

        Start a session like what is done logging into the BUI.  This is not a
        requirement to run REST commands, since the protocol is stateless.
        What is does is set up a cookie session so that some server side
        caching can be done.  If login is used remember to call logout when
        finished.

        :param auth_str: Authorization string (base64)
        """
        self.auth_str = auth_str
        self._authorize()

    def logout(self):
        """Logout of an appliance"""
        result = None
        try:
            result = self.delete("/access/v1", base_path="/api")
        except RestClientError:
            pass

        self.headers.clear()
        self.do_logout = False
        return result

    def islogin(self):
        """return if client is login"""
        return self.do_logout

    @staticmethod
    def mkpath(*args, **kwargs):
        """Make a path?query string for making a REST request

        :cmd_params args: The path part
        :cmd_params kwargs: The query part
        """
        buf = six.StringIO()
        query = "?"
        for arg in args:
            buf.write("/")
            buf.write(arg)
        for k in kwargs:
            buf.write(query)
            if query == "?":
                query = "&"
            buf.write(k)
            buf.write("=")
            buf.write(kwargs[k])
        return buf.getvalue()

    def request(self, path, request, body=None, **kwargs):
        """Make an HTTP request and return the results

        :param path: Path used with the initialized URL to make a request
        :param request: HTTP request type (GET, POST, PUT, DELETE)
        :param body: HTTP body of request
        :key accept: Set HTTP 'Accept' header with this value
        :key base_path: Override the base_path for this request
        :key content: Set HTTP 'Content-Type' header with this value
        """
        out_hdrs = dict.copy(self.headers)
        if kwargs.get("accept"):
            out_hdrs['accept'] = kwargs.get("accept")

        if body:
            if isinstance(body, dict):
                body = json.dumps(body)
            body = body.encode("utf-8")
            out_hdrs['content-length'] = len(body)
        else:
            body = None

        zfssaurl = self._path(path, kwargs.get("base_path"))
        req = urllib.request.Request(zfssaurl, body, out_hdrs)
        req.get_method = lambda: request
        maxreqretries = kwargs.get("maxreqretries", 10)
        retry = 0
        response = None

        LOG.debug('Request: %(request)s %(url)s',
                  {'request': request, 'url': zfssaurl})
        LOG.debug('Out headers: %s', out_hdrs)
        if body and body != '':
            # body may contain chap secret so must be masked
            LOG.debug('Body: %s', strutils.mask_password(body))

        context = None
        if hasattr(ssl, '_create_unverified_context'):
            context = ssl._create_unverified_context()
        else:
            context = None

        while retry < maxreqretries:
            try:
                if context:
                    # only schemes that can be used will be http or https if it
                    # is given in the path variable, or the path will begin
                    # with the REST API location meaning invalid or unwanted
                    # schemes cannot be used
                    response = urllib.request.urlopen(req,  # nosec
                                                      timeout=self.timeout,
                                                      context=context)
                else:
                    response = urllib.request.urlopen(req,  # nosec : see above
                                                      timeout=self.timeout)
            except urllib.error.HTTPError as err:
                if err.code == http_client.NOT_FOUND:
                    LOG.debug('REST Not Found: %s', err.code)
                else:
                    LOG.error('REST Not Available: %s', err.code)

                if err.code == http_client.SERVICE_UNAVAILABLE and \
                   retry < maxreqretries:
                    retry += 1
                    time.sleep(1)
                    LOG.error('Server Busy retry request: %s', retry)
                    continue
                if (err.code == http_client.UNAUTHORIZED or
                    err.code == http_client.INTERNAL_SERVER_ERROR) and \
                   '/access/v1' not in zfssaurl:
                    try:
                        LOG.error('Authorizing request: %(zfssaurl)s '
                                  'retry: %(retry)d.',
                                  {'zfssaurl': zfssaurl, 'retry': retry})
                        self._authorize()
                        req.add_header('x-auth-session',
                                       self.headers['x-auth-session'])
                    except RestClientError:
                        pass
                    retry += 1
                    time.sleep(1)
                    continue

                return RestResult(err=err)

            except urllib.error.URLError as err:
                LOG.error('URLError: %s', err.reason)
                raise RestClientError(-1, name="ERR_URLError",
                                      message=err.reason)

            break

        if (response and
            (response.getcode() == http_client.SERVICE_UNAVAILABLE and
                retry >= maxreqretries)):
            raise RestClientError(response.getcode(), name="ERR_HTTPError",
                                  message="REST Not Available: Disabled")

        return RestResult(response=response)

    def get(self, path, **kwargs):
        """Make an HTTP GET request

        :param path: Path to resource.
        """
        return self.request(path, "GET", **kwargs)

    def post(self, path, body="", **kwargs):
        """Make an HTTP POST request

        :param path: Path to resource.
        :param body: Post data content
        """
        return self.request(path, "POST", body, **kwargs)

    def put(self, path, body="", **kwargs):
        """Make an HTTP PUT request

        :param path: Path to resource.
        :param body: Put data content
        """
        return self.request(path, "PUT", body, **kwargs)

    def delete(self, path, **kwargs):
        """Make an HTTP DELETE request

        :param path: Path to resource that will be deleted.
        """
        return self.request(path, "DELETE", **kwargs)

    def head(self, path, **kwargs):
        """Make an HTTP HEAD request

        :param path: Path to resource.
        """
        return self.request(path, "HEAD", **kwargs)
